# =============================================================================
#     Author: Casey R. Densmore, 12FEB2022
#
#    This file is part of the Airborne eXpendable Buoy Processing System (AXBPS)
#
#    AXBPS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    AXBPS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with AXBPS.  If not, see <https://www.gnu.org/licenses/>.
# =============================================================================


###################################################################################
#                                    IMPORTS                                      #
###################################################################################


import logging

import numpy as np
from scipy import signal, interpolate, optimize, stats
from scipy.io import wavfile #for wav file reading

    
        

###################################################################################
#                         lagging box smooth filter                               #
###################################################################################

def boxsmooth_lag(data,window,startind): #filter is lagging for realtime processing
    outdata = data.copy()
    
    for i in range(startind,len(data)):
        if i < window:
            outdata[i] = np.nanmean(data[0:i+1])
        else:
            outdata[i] = np.nanmean(data[i-window:i+1])
            
    return outdata
    
    
    


###################################################################################
#                         AXCTD PCM DATA FSK DEMODULATION                         #
###################################################################################
    

def demodulate_axctd(pcm, fs, edge_buffer, sos, bitrate, f1, f2, trig1, trig2, Npcm, bit_inset, phase_error, high_bit_scale):
    
    #basic configuration- default options
    # f1 = 400 # bit 1 (mark) = 400 Hz
    # f2 = 800 # bit 0 (space) = 800 Hz
    # bitrate = 800 #symbol rate = 800 Hz
    # bit_inset = 1 #number of points after/before zero crossing where bit identification starts/ends
    # phase_error = 25
    # sos = signal.butter(6, 1200, btype='lowpass', fs=fs, output='sos') #low pass
    # N = int(np.round(fs/bitrate*(1 - phase_error/100))) #first crossing following previous that could be a full bit
    # Npcm = N - 2*bit_inset
    # trig1 = 2*np.pi*np.arange(0,Nn)/fs*f1 #trig term for power calculation
    # trig2 = 2*np.pi*np.arange(0,Nn)/fs*f2
    
    # apply filter to extract FSK data in desired frequency range (<1200 Hz) only
    pcmlow = signal.sosfilt(sos, pcm)
    
    #finding all zero crossings
    pcmsign = np.sign(pcmlow)
    pcmsign[pcmsign == 0] = 1
    zerocrossings = np.where(pcmsign[:-1] != pcmsign[1:])[0]
    
    #ignoring all zero crossings before our starting point
    zerocrossings = np.delete(zerocrossings, np.where(zerocrossings < edge_buffer)) 
    
    #identify zero crossings to separate bits within phase error percentage 
    bit_edges = [zerocrossings[0]]
    prev = zerocrossings[0]
    
    #identifying bit edges
    c = 0 #next available index that could be a bit edge
    while c < len(zerocrossings)-5:
        next_options = zerocrossings[c+1:c+5]
        c += 1 + np.argmin(abs(next_options - (zerocrossings[c] + fs/bitrate)))
        bit_edges.append(zerocrossings[c])
        
    #calculate power at FSK frequencies for each "bit"
    s1 = [] #stores power levels 
    s2 = []
    
    for e in bit_edges[:-1]:
        cdata = pcmlow[e+bit_inset:e+bit_inset+Npcm]
        s1.append(np.abs(np.sum(cdata*np.cos(trig1) + 1j*cdata*np.sin(trig1))))
        s2.append(np.abs(np.sum(cdata*np.cos(trig2) + 1j*cdata*np.sin(trig2)))*high_bit_scale)
        
    next_ind = bit_edges[-1] - 1 #first index to start demodulation on next round
        
    #determine each bit and associated confidence (power of identified freq / power of alternate freq)
    bits = []
    conf = []
    for (p1,p2) in zip(s1,s2):
        conf.append(p2/p1)
        if p1 >= p2:
            bits.append(1)
        else:
            bits.append(0)
            
    return bits, conf, bit_edges, next_ind
            
            
    
###################################################################################
#                           OPTIMIZE DEMOD SCALE FACTOR                           #
###################################################################################

def adjust_scale_factor(confs,scale_factor):
    
    Npts = len(confs)
    confs = np.asarray(confs)
    
    #calculating histogram of confidences
    bin_edges = np.arange(0.0,3,0.01)
    distribution,bin_edges = np.histogram(confs, bins=bin_edges)
    bin_centers = bin_edges[:-1] + np.diff(bin_edges)/2
    
    #calculate cumulative percentage of confidence occurences
    cumpct = 100*np.cumsum(distribution)/Npts
    
    #slope of cumulative percentage function
    cumpctslope = np.array((cumpct[1]-cumpct[0])/(bin_centers[1] - bin_centers[0]))
    cumpctslope = np.append(cumpctslope, (cumpct[2:]-cumpct[:-2])/(bin_centers[2:] - bin_centers[:-2]))
    cumpctslope = np.append(cumpctslope, (cumpct[-1]-cumpct[-2])/(bin_centers[-1] - bin_centers[-2]))
    
    #limiting threshold to cumulative percentages between 30% and 65%
    in_range = [True if cp >= 30 and cp <= 65 else False for cp in cumpct]
    bin_centers = bin_centers[in_range]
    cumpct = cumpct[in_range]
    cumpctslope = cumpctslope[in_range]
    
    #identifying minimum slope and first/last occurences to determine where to cut off bits
    min_slope = np.min(cumpctslope)
    isminslope = np.where(cumpctslope == min_slope)[0]
    new_threshold = np.nanmean([bin_centers[isminslope[0]], bin_centers[isminslope[-1]]])
    
    
    #calcuating new scale factor
    scale_factor /= new_threshold
    
    return scale_factor










    
    


# =============================================================================
#     Author: Casey R. Densmore, 12FEB2022
#
#    AXBPS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    AXBPS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with AXBPS.  If not, see <https://www.gnu.org/licenses/>.
# =============================================================================



import numpy as np
from scipy import signal
from scipy.io import wavfile #for wav file reading

import time as timemodule
from datetime import date

from traceback import print_exc as trace_error

from . import geomag_axbps as gm

#add package directory to path for other file imports
import os, inspect
global cdir_axcp
cdir_axcp = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))



#this could be problematic for debugging but I'm getting tired of the "chunk could not be understood" messages
import warnings
warnings.filterwarnings("ignore")



#reading the audio file
def readAXCPwavfile(inputfile, timerange):
    
    #reading WAV file
    fs, snd = wavfile.read(inputfile)
    
    #if multiple channels, sum them together
    sndshape = np.shape(snd) #array size (tuple)
    ndims = len(sndshape) #number of dimensions
    if ndims == 1: #if one channel, use that
        audiostream = snd
    elif ndims == 2: #if two channels
        #audiostream = np.sum(snd,axis=1) #sum them up
        audiostream = snd[:,0] #use first channel
    else:
        raise Exception('Too many dimensions for an audio file!')
    
    #Normalize amplitude/DC offset of audio signal
    pcm_dc = np.mean(audiostream)
    pcm_ampl = np.max(np.abs(audiostream))
    pcm = (audiostream.astype(np.float) - pcm_dc) / (2 * pcm_ampl)
        
    # downsampling if necessary 
    if fs > 50000: 
        pcm = signal.decimate(pcm, 2)
        fs /= 2
            
    #trimming PCM data to specified range as required
    if timerange[1] > 0: #trim end of profile first to avoid throwing off indicies
        e = int(self.fs*timerange[1])
        self.audiostream = self.audiostream[:e]
    if timerange[0] > 0:
        s = int(self.fs*timerange[0])
        self.audiostream = self.audiostream[s:]
            
        
    return pcm, fs
    
    
#AXCP Processor profile variables: 
#Variables with one point per processing loop (includes profile and noise before (and possibly after) the profile:
# T-time (sec)   CCENV-compass coil signal envelope   PK-peak audio amplitude   FCCDEV-compass coil frequency deviation   
# FROTLP-probe rotation rate (Hz)   FROTDEV-standard deviation of probe rotation rate during each datapoint
#
#Variables with one point per depth datapoint (valid profile data only):
# TIME-time (sec)   DEPTH-depth (m)   TEMP-Temperature (C)   TERR-Temperature error (C)   FTBL-Temperature frequency 
# U_MAG-east/west current velocity oriented to mag north (east > 0)   V_MAG-as in U_MAG but north/south (north > 0)
# U_TRUE/V_TRUE-as in U_MAG/V_MAG but oriented to true north   VERR-velocity error estimate
# ROTF-probe rotation rate (Hz)   ROTFRMS-probe rotation rate RMS deviation
# EFBL- baseline EF frequency corresponding to voltage across AXCP electrodes
# CCBL- baseline frequency of compass coil corresponding to probe orientation (should oscillate at ~16 Hz)
# FEFR, FCCR, VC0A, VC0P, VC0P, VE0A, VE0P, GEFA, GCCA, NINDEP- variables related to velocity calculations (these are necessary to recalculate currents with an updated latitude/longitude)
# ENVCC-envelope of the compass coil signal   ENVCCRMS-RMS deviation of ENVCC
# AREA-estimate of the coil area from compass coil data   AERR-area error estimate PEAK-peak audio amplitude
# W-probe sink rate

class AXCP_Processor:
    
    #importing necessary functions to handle AXCP processing (automatically attaches them to self)
    # from ._AXCP_decode_fxns import (init_AXCP_settings, initialize_AXCP_vars, init_filters, init_constants, first_subsample, second_subsample, calc_current_datapoint, iterate_AXCP_process, refine_spindown_prof, calculate_true_velocities)
    from ._AXCP_decode_fxns import (init_AXCP_settings, initialize_AXCP_vars, init_fft_window, dofft, init_filters, init_constants, first_subsample, second_subsample, calc_current_datapoint, iterate_AXCP_process, refine_spindown_prof, calculate_true_velocities)
    from ._AXCP_convert_fxns import (calc_temp_from_freq, calc_vel_components, calc_currents)

    #initializing current thread (saving variables, reading audio data or contacting/configuring receiver)
    #AXBT settings: fftwindow, minfftratio, minsiglev, triggerfftratio, triggersiglev, tcoeff, zcoeff, flims
    def __init__(self, audiofile, timerange=[0,-1], lat=20, lon=-80, dropdate=date.today(), settings={}):
        
        #reading in WAV file
        self.audiofile = audiofile
        self.audiostream, self.f_s = readAXCPwavfile(audiofile, timerange)
        self.numpoints = len(self.audiostream)
        
        #initialize default settings, override user-specified ones
        self.init_AXCP_settings(settings)     
        
        #updating position, magnetic field components/declination
        self.lat = lat
        self.lon = lon
        self.dropdate = dropdate
        print(os.getcwd())
        self.gm = gm.GeoMag(wmm_filename=os.path.join(cdir_axcp, 'WMM.COF'))
        self.update_position()
        
        #initializing AXCP processor specific vars, as well as filter and conversion coefficients and output profile arrays
        self.initialize_AXCP_vars()
                
        
        
    def update_position(self):
        self.magvar = self.gm.get_params(dlat=self.lat, dlon=self.lon, time=self.dropdate)
        self.fh = self.magvar.bh #always positive
        self.fz = -self.magvar.bz #switches convention so positive is up
        self.dec = self.magvar.dec #positive is East
        
        
        
        
        
    def run(self):
    
        
        self.maxtime = self.numpoints/self.f_s
        
            
        # setting up thread while loop- terminates when user clicks "STOP" or audio file finishes processing
        i = -1
        
        self.keepgoing = True
        
        # initialize state- probe hasn't spun up yet
        self.status = 0
        
        #initialize self.demodbufferstartind
        self.demodbufferstartind = 0
        
        
        #MAIN PROCESSOR LOOP
        while self.keepgoing:
            i += 1
            
                
            #kill threads once time exceeds the max time of the audio file
            #NOTE: need to do this on the cycle before hitting the max time when processing from audio because the WAV file processes faster than the thread can kill itself
            
            #calculating end of next slice of PCM data for signal level calcuation and demodulation
            e = self.demodbufferstartind + self.pointsperloop
            
            if self.numpoints - self.demodbufferstartind < 4*self.f_s: #kill process at file end
                self.keepgoing = False
            
            elif e >= self.numpoints: #terminate loop if at end of file
                e = self.numpoints - 1
                self.keepgoing = False
            
            #add next round of PCM data to buffer for signal calculation and demodulation
            # self.demod_buffer = np.append(self.demod_buffer, self.audiostream[self.demodbufferstartind:e])
            self.demod_buffer = self.audiostream[self.demodbufferstartind:e]
                
            
            print(f"[+] Processing status: {round(100*self.demodbufferstartind/self.numpoints)}%         ", end='\r')
            
            if self.keepgoing: #only process buffer if there is enough data
            
                #demodulating and parsing current batch of AXCTD PCM data
                self.iterate_AXCP_process(e)
    
                #incrementing demod buffer start forward
                self.demodbufferstartind = e 
                    
                    
                    
            #sleeping until ready to process more data 
            timemodule.sleep(0.001) #slight pause to free some resources when processing from audio
            
            
        #after finishing entire profile, refine the spindown point, correct amean, and adjust the profile
        print("[+] Processing status: 100%         ")
        if self.tspinup >= 0 and len(self.TIME) > 0: #spinup detected, valid profile points recorded
            self.refine_spindown_prof()
            self.calculate_true_velocities()
        else:
            print(f"[!] Audio file processed- spinup not detected")
            
            
        
        
        
    


# =============================================================================
#     Author: Casey R. Densmore, 12FEB2022
#
#    This file is part of the Airborne eXpendable Buoy Processing System (AXBPS)
#
#    AXBPS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    AXBPS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with AXBPS.  If not, see <https://www.gnu.org/licenses/>.
# =============================================================================

import numpy as np
from scipy import signal
from scipy.io import wavfile #for wav file reading

import time as timemodule
import datetime as dt

from traceback import print_exc as trace_error

from . import demodulate, parse


#add package directory to path for other file imports
import os, inspect
global cdir_axctd
cdir_axctd = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))


#this could be problematic for debugging but I'm getting tired of the "chunk could not be understood" messages
import warnings
warnings.filterwarnings("ignore")


#reading the audio file
def readAXCTDwavfile(inputfile, timerange):
    
    #reading WAV file
    fs, snd = wavfile.read(inputfile)
    
    #if multiple channels, sum them together
    sndshape = np.shape(snd) #array size (tuple)
    ndims = len(sndshape) #number of dimensions
    if ndims == 1: #if one channel, use that
        audiostream = snd
    elif ndims == 2: #if two channels
        #audiostream = np.sum(snd,axis=1) #sum them up
        audiostream = snd[:,0] #use first channel
    else:
        raise Exception('Too many dimensions for an audio file!')
    
    # Normalize amplitude/DC offset of audio signal
    pcm_dc = np.mean(audiostream)
    pcm_ampl = np.max(np.abs(audiostream))
    pcm = (audiostream.astype(np.float) - pcm_dc) / pcm_ampl
        
    # downsampling if necessary 
    if fs > 50000: 
        pcm = signal.decimate(pcm, 2)
        fs /= 2
                
    #trimming PCM data to specified range as required
    if timerange[1] > 0: #trim end of profile first to avoid throwing off indicies
        e = int(self.fs*timerange[1])
        self.audiostream = self.audiostream[:e]
    if timerange[0] > 0:
        s = int(self.fs*timerange[0])
        self.audiostream = self.audiostream[s:]
            
        
    return pcm, fs
    
    
    
    
    

class AXCTD_Processor:
    
    

    #initializing current thread (saving variables, reading audio data or contacting/configuring receiver)
    #AXBT settings: fftwindow, minfftratio, minsiglev, triggerfftratio, triggersiglev, tcoeff, zcoeff, flims
    def __init__(self, audiofile, timerange=[0,-1], user_settings={}):
        
        #reading in WAV file
        self.audiofile = audiofile
        self.audiostream, self.f_s = readAXCTDwavfile(audiofile, timerange)
        self.numpoints = len(self.audiostream)
        
        
        self.init_default_AXCTD_settings()
        for csetting in user_settings:
            self.settings[csetting] = user_settings[csetting] #overwrite defaults for user specified settings        
            
        #initializing AXCTD processor specific vars
        self.initialize_AXCTD_vars()
        
        #loads variables for AXCTD settings- must be called after initialize_common_vars which writes settings to self.settings
        self.load_AXCTD_settings()
        
        #initializing profile arrays
        self.time = []
        self.r400_prof = []
        self.r7500_prof = []
        self.hexframes = []
        self.depth = []
        self.temperature = []
        self.conductivity = []
        self.salinity = []
        
        
            
            
    def initialize_AXCTD_vars(self):
        
        self.keepgoing = True  # signal connections
        
        self.past_headers = False #when false, program may try to read header data
        self.header1_read = False #notes when each header has been successfully read, to stop trying
        self.header2_read = False
        self.header3_read = False
        
        #temperature lookup table, calibration coefficients
        self.metadata = parse.initialize_axctd_metadata()
        self.metadata['counter_found_2'] = [False] * 72
        self.metadata['counter_found_3'] = [False] * 72
        self.tempLUT = parse.read_temp_LUT(os.path.join(cdir_axctd,'temp_LUT.txt'))
        
        #store powers at different frequencies used to ID profile start
        self.p400 = np.array([])
        self.p7500 = np.array([])
        self.pdead = np.array([])
        self.r400 = np.array([])
        self.r7500 = np.array([])
        self.power_inds = []
        
        self.firstpulse400 = -1 #will store r400 index corresponding to first 400 Hz pulse
        self.profstartind = -1
        self.lastdemodind = -1 #set to -1 to indicate no demodulation has occurred
        self.curpulse = 0
        
        self.firstpointtime = -1
        self.firstpulsetime = -1
        self.status = 0 #0=nothing, 1 = 400 Hz pulses, 2 = 7.5kHz tone
                        
        self.mean7500pwr = np.NaN
                
        #constant settings for analysis
        self.f_s_power = 25 #check 25 times per second aka once per frame
        self.N_power = int(self.f_s/10) #each power calculation is for 0.1 seconds of PCM data
        self.power_smooth_window = 5
        self.d_pcm = int(np.round(self.f_s/self.f_s_power)) #how many points apart to sample power
        
        self.demod_buffer = np.array([]) #buffer (list) for demodulating PCM data
        self.demod_Npad = 100 #how many points to pad on either side of demodulation window (must be larger than window length for low-pass filter in demodulation function)
        self.next_demod_ind = 0
        
        self.high_bit_scale = 1.5 #scale factor for high frequency bit to correct for higher power at low frequencies (will be adjusted to optimize demodulation after reading first header data)
        
        #optional adjustment- make these settings adjustable via AXBPS GUI (shouldn't be necessary though)
        bitrate=800
        bit_inset=1
        phase_error=25
        self.bitrate = bitrate #symbol rate = 800 Hz
        self.bit_inset = bit_inset #number of points after/before zero crossing where bit identification starts/ends
        self.phase_error = phase_error
        N = int(np.round(self.f_s/self.bitrate*(1 - self.phase_error/100))) #length of PCM for bit power test
        self.Npcm = N - 2*self.bit_inset
        
        #buffers
        self.binary_buffer = [] #buffer for demodulated binary data not organized into frames
        self.binary_buffer_inds = [] #pcm indices for each bit start point (used to calculate observation times during frame parsing)
        self.binary_buffer_conf = [] #confidence ratios: used for demodulation debugging/improvement
        self.r400_buffer = []
        self.r7500_buffer = [] #holds 7500 Hz sig lev ratios corresponding to each bit
        
        
        #-1: not processing, 0: no pulses, 1: found pulse, 2: active profile parsing
        self.status = -1 
        
        
    #initializes dict with default settings for AXCTD processor
    #any specified settings will be overwritten after this function is called in __init__
    def init_default_AXCTD_settings(self):
        
        self.settings = {}
        
        self.settings['minr400']          = 2.0         #header pulse power threshold
        self.settings['mindr7500']        = 1.5         #profile tone power threshold
        self.settings['deadfreq']         = 3000        #"quiet" frequency to use as baseline for power calcs
        self.settings['triggerrange']     = [30,-1],    #bounds for how soon to trigger profile after first pulse
        self.settings['mark_space_freqs'] = [400,800]   #mark and space bit frequencies (shouldn't be adjusted)
        self.settings['bitrate']          = 800         #symbol rate (shouldn't be adjusted)
        self.settings['bit_inset']        = 1           #how many PCM points to inset for demod bit identification
        self.settings['phase_error']      = 25          #phase error to consider for zero crossing trackind in demod
        self.settings['usebandpass']      = False       #use bandpass filter instead of lowpass?
        self.settings['refreshrate']      = 2.0         #how many seconds of data to process per loop
        
        #default conversion coefficients to use if headers can't be decoded
        self.settings["zcoeff_axctd"]     = [0.72, 2.76124, -0.000238007, 0]
        self.settings["tcoeff_axctd"]     = [-0.053328, 0.994372, 0.0, 0.0]
        self.settings["ccoeff_axctd"]     = [-0.0622192, 1.04584, 0.0, 0.0]
        
        self.settings["tlims_axctd"]      = [-10,50] #min/max temperature for a good datapoint
        self.settings["slims_axctd"]      = [-1,100] #min/max salinity for a good datapoint
        
        
    #pulls AXCTD settings from dict format into variables attached to AXCTD_Processor class
    def load_AXCTD_settings(self):
        
        #Default settings for processor (tcoeffdefault and ccoeffdefault may vary with AXCTD used)
        # minR400=2.0, mindR7500=1.5, deadfreq=3000, mintimeperloop=2, triggerrange=[30,-1], 
        # mark_space_freqs=[400,800], bitrate=800, bit_inset=1, phase_error=25, use_bandpass=False, 
        # zcoeffdefault=[0.72, 2.76124, -0.000238007, 0], tcoeffdefault=[0,1,0,0], ccoeffdefault=[0,1,0,0]
        
        #settings pulled from AXBPS GUI and passed to AXCTD_Processor threads
        # settingstopull = ["minr400", "mindr7500", "deadfreq", "refreshrate", "mark_space_freqs", "usebandpass", "zcoeff_axctd", "tcoeff_axctd", "ccoeff_axctd"]
        
        self.minpointsperloop = int(self.settings['refreshrate']*self.f_s) #how many PCM datapoints AXCTDprocessor handles per loop
        
        #signal to noise ratio settings
        self.minR400 = self.settings['minr400'] #threshold to ID first 400 Hz pulse
        self.minR400_inprof = self.settings['minr400']/2 #threshold of 400 Hz signal in profile for good data
        self.mindR7500 = self.settings['mindr7500'] #threshold to ID profile start by 7.5 kHz tone
        self.mindR7500_inprof = self.mindR7500/2 #threshold of 7.5 kHz tone power for good profile datapoint
        self.deadfreq = self.settings['deadfreq'] #frequency to use as "data-less" control to normalize signal levels
        
        #default depth/temperature/conductivity conversion coefficients
        self.zcoeff = self.settings['zcoeff_axctd']
        self.tcoeff = self.settings['tcoeff_axctd']
        self.ccoeff = self.settings['ccoeff_axctd']
        
        #max ranges for good temperature and salinity values
        self.tlims = self.settings['tlims_axctd']
        self.slims = self.settings['slims_axctd']
        
        #demodulator configuration
        self.f1 = self.settings['mark_space_freqs'][0] # bit 1 (mark) = 400 Hz
        self.f2 = self.settings['mark_space_freqs'][1] # bit 0 (space) = 800 Hz
        
        #recalculating trig terms
        self.trig1 = 2*np.pi*np.arange(0,self.Npcm)/self.f_s*self.f1 #trig term for power calculation
        self.trig2 = 2*np.pi*np.arange(0,self.Npcm)/self.f_s*self.f2
        
        #index 0: earliest time AXBT will trigger after 400 Hz pulse in seconds (default 30 sec)
        #index 1: time AXCTD will autotrigger without 7.5kHz signal (set to -1 to never trigger profile)
        triggerrange = [30,-1] #TODO: Add a setting for this in AXBPS GUI 
        self.triggerrange = triggerrange
        
        #filter to be applied to PCM data before demodulation
        if self.settings['usebandpass']:
            self.sos_filter = signal.butter(6, [100,1200], btype='bandpass', fs=self.f_s, output='sos') #low pass
        else:
            self.sos_filter = signal.butter(6, 1200, btype='lowpass', fs=self.f_s, output='sos') #low pass
            
        #trig terms for power calculations
        self.theta400 = 2*np.pi*np.arange(0,self.N_power)/self.f_s*400 #400 Hz (main pulse)
        self.theta7500 = 2*np.pi*np.arange(0,self.N_power)/self.f_s*7500 #400 Hz (main pulse)
        self.thetadead = 2*np.pi*np.arange(0,self.N_power)/self.f_s*self.deadfreq #400 Hz (main pulse)
        
        
        
        
    def run(self):
    
        
        self.maxtime = self.numpoints/self.f_s
        
            
        # setting up thread while loop- terminates when user clicks "STOP" or audio file finishes processing
        i = -1
        
        self.status = 0
        
        #initialize self.demodbufferstartind
        self.demodbufferstartind = 0
        
        
        #MAIN PROCESSOR LOOP
        while self.keepgoing:
            i += 1

            
                
                
            #kill threads once time exceeds the max time of the audio file
            #NOTE: need to do this on the cycle before hitting the max time when processing from audio because the WAV file processes faster than the thread can kill itself
            
            #calculating end of next slice of PCM data for signal level calcuation and demodulation
            e = self.demodbufferstartind + self.minpointsperloop
            
            if self.numpoints - self.demodbufferstartind < 4*self.N_power: #kill process at file end
                self.keepgoing = False
                print("[+] Processing status: 100%")
            
            elif e >= self.numpoints: #terminate loop if at end of file
                e = self.numpoints - 1
            
            #add next round of PCM data to buffer for signal calculation and demodulation
            # self.demod_buffer = np.append(self.demod_buffer, self.audiostream[self.demodbufferstartind:e])
            self.demod_buffer = self.audiostream[self.demodbufferstartind:e]
                
            
            print(f"[+] Processing status: {round(100*self.demodbufferstartind/self.numpoints)}%         ", end='\r')
            
            if self.keepgoing: #only process buffer if there is enough data
                oldstatus = self.status #track status change to emit triggered signal when necessary
                
                #demodulating and parsing current batch of AXCTD PCM data
                data = self.iterate_AXCTD_process(e)
                
                if len(data) > 1: #retrieved good profile data
                    self.time.extend(data[1])
                    self.r400_prof.extend(data[2])
                    self.r7500_prof.extend(data[3])
                    self.depth.extend(data[4])
                    self.temperature.extend(data[5])
                    self.conductivity.extend(data[6])
                    self.salinity.extend(data[7])
                    self.hexframes.extend(data[8])
                
                    
                #increment demod buffer index forward
                if self.status > 0: #if active demodulation occuring, increment to start of next bit, corrected for padding
                    if self.next_demod_ind > self.demod_Npad:
                        self.demodbufferstartind += self.next_demod_ind - self.demod_Npad
                    else:
                        self.demodbufferstartind += self.f_s/self.bitrate #skip one bit, keep going
                else: #signal level calcs only, increment buffer by pointsperloop
                    self.demodbufferstartind = e 
                    
                    
                    
            #sleeping until ready to process more data 
            timemodule.sleep(0.001) #slight pause to free some resources when processing from audio
                
            
            
    
                    
                    
    #this function is called once per loop of the AXCTD Processor and demodulates/parses as much data as is available in the respective buffers of PCM data and unparsed bits, returning a list of data to the AXCTD_Processor loop to be saved
    def iterate_AXCTD_process(self, e): 

        #getting signal levels at 400 Hz, 7500 Hz, and dead frequency
        #we don't have to calculate all three every time but need to calculate at least two of them every
        #time and the logic/code complexity to save computing an unnecessary one isn't worth the computing
        #power saved
        #sampling interval = sampling frequency / power sampling frequency
        #calculating signal levels at 400 Hz, 7500 Hz, and dead frequency (default 3000 Hz)
        
        pstartind = len(self.power_inds)
        
        self.power_inds.extend([ind for ind in range(self.demodbufferstartind, e-self.N_power, self.d_pcm)])
        for cind in self.power_inds[pstartind:]:
            bufferind = cind - self.demodbufferstartind
            cdata = self.demod_buffer[bufferind:bufferind+self.N_power]
            
            self.p400 = np.append(self.p400, np.abs(np.sum(cdata*np.cos(self.theta400) + 1j*cdata*np.sin(self.theta400))))
            self.p7500 = np.append(self.p7500, np.abs(np.sum(cdata*np.cos(self.theta7500) + 1j*cdata*np.sin(self.theta7500))))
            self.pdead = np.append(self.pdead, np.abs(np.sum(cdata*np.cos(self.thetadead) + 1j*cdata*np.sin(self.thetadead))))
                    
        #smoothing signal levels, calculating R400/R7500
        self.p400 = demodulate.boxsmooth_lag(self.p400, self.power_smooth_window, pstartind)
        self.p7500 = demodulate.boxsmooth_lag(self.p7500, self.power_smooth_window, pstartind)
        self.pdead = demodulate.boxsmooth_lag(self.pdead, self.power_smooth_window, pstartind)
        self.r400 = np.append(self.r400, np.log10(self.p400[pstartind:]/self.pdead[pstartind:]))
        self.r7500 = np.append(self.r7500, np.log10(self.p7500[pstartind:]/self.pdead[pstartind:]))
        
        
        #look for 400 Hz pulse if it hasn't been discovered yet
        if self.status == 0:
            matchpoints = np.where(self.r400[pstartind:] >= self.minR400)
            if len(matchpoints[0]) > 0: #found a match!
                self.firstpulse400 = self.power_inds[pstartind:][matchpoints[0][0]] #getting index in original PCM data
                self.firstpulsetime = self.firstpulse400/self.f_s
                self.status = 1
            
        
        #if pulse discovered, demodulate data to bitstream AND check 7500 Hz power
        if self.status >= 1:
            
            #calculate 7500 Hz lower power (in range 4.5-5.5 seconds after first 400 Hz pulse start)
            #last power index is at least 5.5 seconds after 400 Hz pulse detected
            if self.power_inds[-1] >= self.firstpulse400 + int(self.f_s*5.5) and np.isnan(self.mean7500pwr): 
                #mean signal level at 7500 Hz between 4.5 and 5.5 sec after 400 Hz pulse
                pwr_ind_array = np.asarray(self.power_inds)
                s7500ind = np.argmin(np.abs(self.firstpulse400 + int(self.f_s*4.5) - pwr_ind_array))
                e7500ind = np.argmin(np.abs(self.firstpulse400 + int(self.f_s*5.5) - pwr_ind_array))
                self.mean7500pwr = np.nanmean(self.r7500[s7500ind:e7500ind])
            
            #get 7500 Hz power and update status (and bitstream) as necessary 
            #(only bother if time since 400 Hz pulse exceeds the minimum time for profile start)
            if self.power_inds[-1] > self.firstpulse400 + int(self.triggerrange[0]*self.f_s):
                if not np.isnan(self.mean7500pwr) and self.status == 1:
                    matchpoints = np.where(self.r7500[pstartind:] - self.mean7500pwr >= self.mindR7500)
                    if len(matchpoints[0]) > 0: #if power threshold is exceeded
                        self.profstartind = self.power_inds[pstartind:][matchpoints[0][0]]
                        self.status = 2
                #if the latest trigger time setting has been exceeded
                elif self.triggerrange[1] > 0 and self.power_inds[-1] >= self.firstpulse400 + int(self.f_s*self.triggerrange[1]):
                    self.profstartind = self.firstpulse400 + int(self.f_s*self.triggerrange[1])
                    self.status = 2
                if self.profstartind > 0 and self.firstpointtime <= 0:
                    self.firstpointtime = self.profstartind/self.f_s
                    
            #demodulate to bitstream and append bits to buffer
            curbits, conf, bit_edges, self.next_demod_ind = demodulate.demodulate_axctd(self.demod_buffer, self.f_s, self.demod_Npad, self.sos_filter, self.bitrate, self.f1, self.f2, self.trig1, self.trig2, self.Npcm, self.bit_inset, self.phase_error, self.high_bit_scale)
            
            self.binary_buffer.extend(curbits) #buffer for demodulated binary data not organized into frames
            
            new_bit_inds = [be + self.demodbufferstartind for be in bit_edges]
            self.binary_buffer_inds.extend(new_bit_inds)
            self.binary_buffer_conf.extend(conf)
            
            
            #array of profile signal levels to go with other data 
            recent_r7500 = self.r7500[pstartind:]
            recent_r400 = self.r400[pstartind:]
            recent_pwrinds = self.power_inds[pstartind:]
            
            new_r400 = [recent_r400[np.argmin(np.abs(recent_pwrinds - ci))] for ci in new_bit_inds]
            self.r400_buffer.extend(new_r400)
            
            new_r7500 = [recent_r7500[np.argmin(np.abs(recent_pwrinds - ci))]-self.mean7500pwr for ci in new_bit_inds]
            self.r7500_buffer.extend(new_r7500)
            
            
        #attempting to read headers for conversion coefficients and AXCTD metadata
        if self.status >= 1 and not self.past_headers:
        
            #seeing if enough time has passed since first pulse to contain 2nd or 3rd header data
            #pulse length: 1.8 sec, header length: 2.88 sec, gap period (first 2 pulses): 5 sec
            #total pulse cycle ~= 9.68 sec (assume 9-10 sec)
            
            headerdata = [None,None]
            
            firstbin = self.binary_buffer_inds[0]
            lastbin = self.binary_buffer_inds[-1]
            cbufferindarray = np.asarray(self.binary_buffer_inds)
            
            #first header should start around 1.8 sec and end around 3.7 seconds
            #only processing a small margin within that to be sure we are only capturing 1 sec of header
            p1headerstartpcm = self.firstpulse400 + int(self.f_s*2.3)
            p1headerendpcm = self.firstpulse400 + int(self.f_s*3.3)
            
            #second header should start around 11.48 sec and end around 14.36 seconds
            p2headerstartpcm = self.firstpulse400 + int(self.f_s*10.5) #capture 1 sec of pulse
            p2headerendpcm = self.firstpulse400 + int(self.f_s*14.8) #~half second margin on backend
            
            #third header should start around 21.16 sec and end around 24.04 seconds
            p3headerstartpcm = self.firstpulse400 + int(self.f_s*20) #same margins as header 2
            p3headerendpcm =  self.firstpulse400 + int(self.f_s*24.5)
            
            #establishing signal amplitude ratio cutoff based on first header to improve demodulation
            if firstbin <= p1headerstartpcm and lastbin >= p1headerendpcm and not self.header1_read: 
                
                #determining binary data start/end index (adding extra 0.5 sec of data if available)
                p1startind = np.where(cbufferindarray >= p1headerstartpcm - int(self.f_s*0.5))[0][0]
                p1endind = np.where(cbufferindarray <= p1headerendpcm + int(self.f_s*0.5))[0][-1]
                
                #pulling confidence ratios from the header and recalculating optimal high bit scale
                header_confs = self.binary_buffer_conf[p1startind:p1endind]
                self.high_bit_scale = demodulate.adjust_scale_factor(header_confs, self.high_bit_scale)
                self.header1_read = True
                
            
            #trying to process second header
            if firstbin <= p2headerstartpcm and lastbin >= p2headerendpcm and not self.header2_read: 
                
                #determining binary data start/end index (adding extra 0.5 sec of data if available)
                p2startind = np.where(cbufferindarray >= p2headerstartpcm - int(self.f_s*0.5))[0][0]
                p2endind = np.where(cbufferindarray <= p2headerendpcm + int(self.f_s*0.5))[0][-1]
                
                #pulling header data from pulse
                header_bindata = parse.trim_header(self.binary_buffer[p2startind:p2endind])
                
                if len(header_bindata) >= 72*32: #must contain full header
                
                    #parsing and converting header information
                    headerdata[0] = parse.parse_header(header_bindata)
                    self.header2_read = True #read complete for 2nd header transmission
                    
            #trying to process third header
            if firstbin <= p3headerstartpcm and lastbin >= p3headerendpcm and not self.header3_read: 
                
                #determining binary data start/end index (adding extra 0.5 sec of data if available)
                p3startind = np.where(cbufferindarray >= p3headerstartpcm - int(self.f_s*0.5))[0][0]
                p3endind = np.where(cbufferindarray <= p3headerendpcm + int(self.f_s*0.5))[0][-1]
                
                #pulling header data from pulse
                header_bindata = parse.trim_header(self.binary_buffer[p3startind:p3endind])
                
                if len(header_bindata) >= 72*32: #must contain full header
                
                    #parsing and converting header information
                    headerdata[1] = parse.parse_header(header_bindata)
                    self.header3_read = True #read complete for 3rd header transmission
                    
                    
                
            #incorporating AXCTD header info into profile metadata
            coeffs = ['t','c','z']
            other_data = ['serial_no','probe_code','max_depth','misc']
            for i,header in enumerate(headerdata):
                
                if header is not None:
                    
                    self.metadata[f'frame_data_{i+2}'] = header['frame_data']
                    self.metadata[f'counter_found_{i+2}'] = header['counter_found']
                    
                    for coeff in coeffs: #incorporating coefficients (coeff, coeff_valid, coeff_hex)
                        for ci in range(4):
                            if header[coeff + 'coeff_valid'][ci]:
                                self.metadata[coeff + 'coeff'][ci] = header[coeff + 'coeff'][ci]
                                self.metadata[coeff + 'coeff_hex'][ci] = header[coeff + 'coeff_hex'][ci]
                                self.metadata[coeff + 'coeff_valid'][ci] = True
                    
                    for key in other_data: #incorporating other profile metadata
                        if header[key] is not None and self.metadata[key] is None:
                            self.metadata[key] = header[key]
                    
                            
            
            #if updated headers included, then try to update coefficients
            if headerdata[0] is not None or headerdata[1] is not None: 
                if sum(self.metadata['tcoeff_valid']) == 4:
                    self.tcoeff = self.metadata['tcoeff']
                if sum(self.metadata['ccoeff_valid']) == 4:
                    self.ccoeff = self.metadata['ccoeff']
                if sum(self.metadata['tcoeff_valid']) == 4:
                    self.zcoeff = self.metadata['zcoeff']
            
                    
        pass_empty = True
        
        if self.status == 2: #parsing bitstream into frames and calculating updated profile data
            
            self.past_headers = True
            
            #cutting off all data before profile initiation
            if self.binary_buffer_inds[0] <= self.profstartind:
                firstind = np.where(np.asarray(self.binary_buffer_inds) > self.profstartind)[0][0]
                self.binary_buffer = self.binary_buffer[firstind:]
                self.binary_buffer_inds = self.binary_buffer_inds[firstind:]
                self.binary_buffer_conf = self.binary_buffer_conf[firstind:]
                self.r400_buffer = self.r400_buffer[firstind:]
                self.r7500_buffer = self.r7500_buffer[firstind:]
            
            #calculting times corresponding to each bit
            binbufftimes = (np.asarray(self.binary_buffer_inds) - self.profstartind)/self.f_s
                
            #parsing data into frames
            hexframes, times, depths, temps, conds, psals, r400, r7500, next_buffer_ind = parse.parse_bitstream_to_profile(self.binary_buffer, binbufftimes, self.r400_buffer, self.r7500_buffer, self.tempLUT, self.tcoeff, self.ccoeff, self.zcoeff)
                        
            #rounding data and appending to lists
            times = np.round(np.asarray(times) + self.firstpointtime, 2)
            depths = np.round(depths,2)
            temps = np.round(temps,2)
            conds = np.round(conds,2)
            psals = np.round(psals,2)
            r400 = np.round(r400,2)
            r7500 = np.round(r7500,2)
            
            
            is_good = [True] * len(times)
            
            #if R400, dR7500, temp, or psal are outside of preset bounds, exclude datapoint
            for i,p in enumerate(r7500):
                if p < self.mindR7500_inprof or r400[i] < self.minR400_inprof or temps[i] < self.tlims[0] or temps[i] > self.tlims[1] or psals[i] < self.slims[0] or psals[i] > self.slims[1]:
                    is_good[i] = False
                    
            times = times[is_good]
            depths = depths[is_good]
            temps = temps[is_good]
            conds = conds[is_good]
            psals = psals[is_good]
            r400 = r400[is_good]
            r7500 = r7500[is_good]
            
            
            
            #identifying and removing spikes
            if len(temps) > 0:
                is_good = [True] * len(times)
                
                #median and percentile value based thresholds
                thresh = 10
                pct_offset = 35
                T_median = np.percentile(temps,50)
                T_low_diff_thresh = T_median - thresh*(T_median - np.percentile(temps,50-pct_offset))
                T_high_diff_thresh = T_median + thresh*(np.percentile(temps,50+pct_offset) - T_median)
                S_median = np.percentile(psals,50)
                S_low_diff_thresh = S_median - thresh*(S_median - np.percentile(psals,50-pct_offset))
                S_high_diff_thresh = S_median + thresh*(np.percentile(psals,50+pct_offset) - S_median)
                for i,_ in enumerate(r7500):
                    if temps[i] < T_low_diff_thresh or temps[i] > T_high_diff_thresh or psals[i] < S_low_diff_thresh or psals[i] > S_high_diff_thresh:
                        is_good[i] = False
                        
                times = times[is_good]
                depths = depths[is_good]
                temps = temps[is_good]
                conds = conds[is_good]
                psals = psals[is_good]
                r400 = r400[is_good]
                r7500 = r7500[is_good]
                
                if len(temps) > 0:
                    data = [self.status, times, r400, r7500, depths, temps, conds, psals, hexframes] 
                    pass_empty = False
                    
                
                
            #removing parsed data from binary buffer
            self.binary_buffer = self.binary_buffer[next_buffer_ind:]
            self.binary_buffer_inds = self.binary_buffer_inds[next_buffer_ind:]
            self.r400_buffer = self.r400_buffer[next_buffer_ind:]
            self.r7500_buffer = self.r7500_buffer[next_buffer_ind:]
            
        
        if pass_empty:
            data = [self.status]
                
        return data


    


# geomag.py
# by Christopher Weiss cmweiss@gmail.com

# Adapted from the geomagc software and World Magnetic Model of the NOAA
# Satellite and Information Service, National Geophysical Data Center
# http://www.ngdc.noaa.gov/geomag/WMM/DoDWMM.shtml
#
# Suggestions for improvements are appreciated.

# USAGE:
#
# >>> gm = geomag.GeoMag("WMM.COF")
# >>> mag = gm.GeoMag(80,0)
# >>> mag.dec
# -6.1335150785195536
# >>>

import math, os, unittest
from datetime import date, datetime


#USE EXAMPLE
# gm = GeoMag(wmm_filename='WMM.COF')
# magvar = gm.get_params(dlat=29.150,dlon=-88.140)
# declination: magvar.dec (degrees, E > 0)
# inclination: magvar.dip (degrees, down > 0)
# total field strength: magvar.ti (nT)
# Fh/Fx/Fy/Fz: magvar.bh/bx/by/bz (nT) (** 1 nT = 10**-5 gauss)


class MagParameters:
    
    def __init__(self, dec, dip, ti, bh, bx, by, bz, lat, lon, alt, time):
    
        self.dec = dec
        self.dip = dip
        self.ti = ti
        self.bh = bh
        self.bx = bx
        self.by = by
        self.bz = bz
        self.lat = lat
        self.lon = lon
        self.alt = alt
        self.time = time
        if self.lon > 180:
            self.lon -= 360
    
    def __repr__(self):
        ns = 'N' if self.lat >= 0 else 'S'
        ew = 'E' if self.lon >= 0 else 'W'
        return f"Magnetic Parameters for {abs(self.lat)} {ns}, {abs(self.lon)} {ew}, {self.alt} m, {datetime.strftime(self.time, '%Y %b %d')}"
        

class GeoMag:

    def get_params(self, dlat, dlon, h=0, time=date.today()): # latitude (decimal degrees), longitude (decimal degrees), altitude (feet), date
        #time = date('Y') + date('z')/365
        date_ref = time
        time = time.year+((time - date(time.year,1,1)).days/365.0)
        alt = h/3280.8399

        otime = oalt = olat = olon = -1000.0

        dt = time - self.epoch
        glat = dlat
        glon = dlon
        rlat = math.radians(glat)
        rlon = math.radians(glon)
        srlon = math.sin(rlon)
        srlat = math.sin(rlat)
        crlon = math.cos(rlon)
        crlat = math.cos(rlat)
        srlat2 = srlat*srlat
        crlat2 = crlat*crlat
        self.sp[1] = srlon
        self.cp[1] = crlon

        #/* CONVERT FROM GEODETIC COORDS. TO SPHERICAL COORDS. */
        if (alt != oalt or glat != olat):
            q = math.sqrt(self.a2-self.c2*srlat2)
            q1 = alt*q
            q2 = ((q1+self.a2)/(q1+self.b2))*((q1+self.a2)/(q1+self.b2))
            ct = srlat/math.sqrt(q2*crlat2+srlat2)
            st = math.sqrt(1.0-(ct*ct))
            r2 = (alt*alt)+2.0*q1+(self.a4-self.c4*srlat2)/(q*q)
            r = math.sqrt(r2)
            d = math.sqrt(self.a2*crlat2+self.b2*srlat2)
            ca = (alt+d)/r
            sa = self.c2*crlat*srlat/(r*d)

        if (glon != olon):
            for m in range(2,self.maxord+1):
                self.sp[m] = self.sp[1]*self.cp[m-1]+self.cp[1]*self.sp[m-1]
                self.cp[m] = self.cp[1]*self.cp[m-1]-self.sp[1]*self.sp[m-1]

        aor = self.re/r
        ar = aor*aor
        br = bt = bp = bpp = 0.0
        for n in range(1,self.maxord+1):
            ar = ar*aor
            
            #for (m=0,D3=1,D4=(n+m+D3)/D3;D4>0;D4--,m+=D3):
            m=0
            D3=1
            #D4=(n+m+D3)/D3
            D4=(n+m+1)
            while D4>0:

        # /*
                # COMPUTE UNNORMALIZED ASSOCIATED LEGENDRE POLYNOMIALS
                # AND DERIVATIVES VIA RECURSION RELATIONS
        # */
                if (alt != oalt or glat != olat):
                    if (n == m):
                        self.p[m][n] = st * self.p[m-1][n-1]
                        self.dp[m][n] = st*self.dp[m-1][n-1]+ct*self.p[m-1][n-1]

                    elif (n == 1 and m == 0):
                        self.p[m][n] = ct*self.p[m][n-1]
                        self.dp[m][n] = ct*self.dp[m][n-1]-st*self.p[m][n-1]

                    elif (n > 1 and n != m):
                        if (m > n-2):
                            self.p[m][n-2] = 0
                        if (m > n-2):
                            self.dp[m][n-2] = 0.0
                        self.p[m][n] = ct*self.p[m][n-1]-self.k[m][n]*self.p[m][n-2]
                        self.dp[m][n] = ct*self.dp[m][n-1] - st*self.p[m][n-1]-self.k[m][n]*self.dp[m][n-2]

        # /*
                # TIME ADJUST THE GAUSS COEFFICIENTS
        # */
                if (time != otime):
                    self.tc[m][n] = self.c[m][n]+dt*self.cd[m][n]
                    if (m != 0):
                        self.tc[n][m-1] = self.c[n][m-1]+dt*self.cd[n][m-1]

        # /*
                # ACCUMULATE TERMS OF THE SPHERICAL HARMONIC EXPANSIONS
        # */
                par = ar*self.p[m][n]
                
                if (m == 0):
                    temp1 = self.tc[m][n]*self.cp[m]
                    temp2 = self.tc[m][n]*self.sp[m]
                else:
                    temp1 = self.tc[m][n]*self.cp[m]+self.tc[n][m-1]*self.sp[m]
                    temp2 = self.tc[m][n]*self.sp[m]-self.tc[n][m-1]*self.cp[m]

                bt = bt-ar*temp1*self.dp[m][n]
                bp = bp + (self.fm[m] * temp2 * par)
                br = br + (self.fn[n] * temp1 * par)
        # /*
                    # SPECIAL CASE:  NORTH/SOUTH GEOGRAPHIC POLES
        # */
                if (st == 0.0 and m == 1):
                    if (n == 1):
                        self.pp[n] = self.pp[n-1]
                    else:
                        self.pp[n] = ct*self.pp[n-1]-self.k[m][n]*self.pp[n-2]
                    parp = ar*self.pp[n]
                    bpp = bpp + (self.fm[m]*temp2*parp)
                    
                D4=D4-1
                m=m+1

        if (st == 0.0):
            bp = bpp
        else:
            bp = bp/st
        # /*
            # ROTATE MAGNETIC VECTOR COMPONENTS FROM SPHERICAL TO
            # GEODETIC COORDINATES
        # */
        bx = -bt*ca-br*sa
        by = bp
        bz = bt*sa-br*ca
        # /*
            # COMPUTE DECLINATION (DEC), INCLINATION (DIP) AND
            # TOTAL INTENSITY (TI)
        # */
        bh = math.sqrt((bx*bx)+(by*by))
        ti = math.sqrt((bh*bh)+(bz*bz))
        dec = math.degrees(math.atan2(by,bx))
        dip = math.degrees(math.atan2(bz,bh))
        # /*
            # COMPUTE MAGNETIC GRID VARIATION IF THE CURRENT
            # GEODETIC POSITION IS IN THE ARCTIC OR ANTARCTIC
            # (I.E. GLAT > +55 DEGREES OR GLAT < -55 DEGREES)

            # OTHERWISE, SET MAGNETIC GRID VARIATION TO -999.0
        # */
        gv = -999.0
        if (math.fabs(glat) >= 55.):
            if (glat > 0.0 and glon >= 0.0):
                gv = dec-glon
            if (glat > 0.0 and glon < 0.0):
                gv = dec+math.fabs(glon);
            if (glat < 0.0 and glon >= 0.0):
                gv = dec+glon
            if (glat < 0.0 and glon < 0.0):
                gv = dec-math.fabs(glon)
            if (gv > +180.0):
                gv = gv - 360.0
            if (gv < -180.0):
                gv = gv + 360.0
                
        
        return MagParameters(dec, dip, ti, bh, bx, by, bz, glat, glon, alt, date_ref)
        
        
        

    def __init__(self, wmm_filename=None):
        if not wmm_filename:
            wmm_filename = os.path.join(os.path.dirname(__file__), 'WMM.COF')
        wmm=[]
        with open(wmm_filename) as wmm_file:
            for line in wmm_file:
                linevals = line.strip().split()
                if len(linevals) == 3:
                    self.epoch = float(linevals[0])
                    self.model = linevals[1]
                    self.modeldate = linevals[2]
                elif len(linevals) == 6:
                    linedict = {'n': int(float(linevals[0])),
                    'm': int(float(linevals[1])),
                    'gnm': float(linevals[2]),
                    'hnm': float(linevals[3]),
                    'dgnm': float(linevals[4]),
                    'dhnm': float(linevals[5])}
                    wmm.append(linedict)

        z = [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0]
        self.maxord = self.maxdeg = 12
        self.tc = [z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13]]
        self.sp = z[0:14]
        self.cp = z[0:14]
        self.cp[0] = 1.0
        self.pp = z[0:13]
        self.pp[0] = 1.0
        self.p = [z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14]]
        self.p[0][0] = 1.0
        self.dp = [z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13]]
        self.a = 6378.137
        self.b = 6356.7523142
        self.re = 6371.2
        self.a2 = self.a*self.a
        self.b2 = self.b*self.b
        self.c2 = self.a2-self.b2
        self.a4 = self.a2*self.a2
        self.b4 = self.b2*self.b2
        self.c4 = self.a4 - self.b4

        self.c = [z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14]]
        self.cd = [z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14],z[0:14]]
        
        for wmmnm in wmm:
            m = wmmnm['m']
            n = wmmnm['n']
            gnm = wmmnm['gnm']
            hnm = wmmnm['hnm']
            dgnm = wmmnm['dgnm']
            dhnm = wmmnm['dhnm']
            if (m <= n):
                self.c[m][n] = gnm
                self.cd[m][n] = dgnm
                if (m != 0):
                    self.c[n][m-1] = hnm
                    self.cd[n][m-1] = dhnm

        #/* CONVERT SCHMIDT NORMALIZED GAUSS COEFFICIENTS TO UNNORMALIZED */
        self.snorm = [z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13]]
        self.snorm[0][0] = 1.0
        self.k = [z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13],z[0:13]]
        self.k[1][1] = 0.0
        self.fn = [0.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0,10.0,11.0,12.0,13.0]
        self.fm = [0.0,1.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0,10.0,11.0,12.0]
        for n in range(1,self.maxord+1):
            self.snorm[0][n] = self.snorm[0][n-1]*(2.0*n-1)/n
            j=2.0
            #for (m=0,D1=1,D2=(n-m+D1)/D1;D2>0;D2--,m+=D1):
            m=0
            D1=1
            D2=(n-m+D1)/D1
            while (D2 > 0):
                self.k[m][n] = (((n-1)*(n-1))-(m*m))/((2.0*n-1)*(2.0*n-3.0))
                if (m > 0):
                    flnmj = ((n-m+1.0)*j)/(n+m)
                    self.snorm[m][n] = self.snorm[m-1][n]*math.sqrt(flnmj)
                    j = 1.0
                    self.c[n][m-1] = self.snorm[m][n]*self.c[n][m-1]
                    self.cd[n][m-1] = self.snorm[m][n]*self.cd[n][m-1]
                self.c[m][n] = self.snorm[m][n]*self.c[m][n]
                self.cd[m][n] = self.snorm[m][n]*self.cd[m][n]
                D2=D2-1
                m=m+D1

class GeoMagTest(unittest.TestCase):

    d1=date(2015,1,1)
    d2=date(2017,7,2)
    
    test_values = (
        # date, alt, lat, lon, var
        (d1, 0, 80, 0,  -3.85),
        (d1, 0, 0, 120, 0.57),
        (d1, 0, -80, 240,  69.81),
        (d1, 328083.99, 80, 0, -4.27),
        (d1, 328083.99, 0, 120, 0.56),
        (d1, 328083.99, -80, 240, 69.22),
        (d2, 0, 80, 0, -2.75),
        (d2, 0, 0, 120, 0.32),
        (d2, 0, -80, 240, 69.58),
        (d2, 328083.99, 80, 0, -3.17),
        (d2, 328083.99, 0, 120, 0.32),
        (d2, 328083.99, -80, 240, 69.00),
    )
    
    def test_declination(self):
        gm = GeoMag()
        for values in self.test_values:
            calcval=gm.GeoMag(values[2], values[3], values[1], values[0])
            self.assertAlmostEqual(values[4], calcval.dec, 2, 'Expected %s, result %s' % (values[4], calcval.dec))

if __name__ == '__main__':
    unittest.main()
    


# =============================================================================
#     Author: Casey R. Densmore, 12FEB2022
#
#    AXBPS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    AXBPS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with AXBPS.  If not, see <https://www.gnu.org/licenses/>.
# =============================================================================

# =============================================================================
#   This code is translated/updated from the MATLAB xcpdsp.m script written by 
#       John Dunlap, University of Washington APL, 4 November 2009
# =============================================================================

import numpy as np


#conversion: coefficients=C,  D_out = C[0] + C[1]*D_in + C[2]*D_in^2 + C[3]*D_in^3 + ...
def dataconvert(data_in,coefficients):
    
    datatype = 1 #integer or float
    if type(data_in) == list:
        datatype = 2
    elif type(data_in) == np.ndarray: #numpy array
        dataype = 3
        
    if datatype == 1:
        data_in = [data_in]
        
    output = []
    for cur_data_in in data_in:
        cur_output = 0
        for (i,c) in enumerate(coefficients):
            cur_output += c*cur_data_in**i
        output.append(cur_output)
        
    if datatype == 1: #convert back from list to int/float
        output = output[0]
    elif datatype == 3: #convert to np array
        output = np.asarray(output)
            
    return output
    
    


def calc_temp_from_freq(self, freq, depth):
    teres = 1.0/(4.4*freq*self.tcalcap)-self.tcalrs
    if teres>0:
        ln = np.log(teres) 
    else:
        ln = np.NaN 

    temp=1.0/(self.tcal[0]+ln*(self.tcal[1]+ln*(self.tcal[2]+ ln*self.tcal[3]))) - 273.15
    
    # adjust temperature to match mendo.rt processing
    temp = temp + self.tcor[0] + self.tcor[1] * depth
    
    return temp
    



def calc_vel_components(self, fcss, fess, tss):

    # tz to get rotation period
    # interpolate fcss to get positive zero crossing times, tz
    j = np.where(np.isfinite(fcss))[0]
    x = fcss - np.mean(fcss[j]) 
    r = np.ones(len(x))
    r[x < 0] = -1
    jp = np.where(np.diff(r) > 0)[0]
    dxjp =   x[jp+1] -   x[jp]
    dtjp = tss[jp+1] - tss[jp]
    tzp = tss[jp] - x[jp] * dtjp / dxjp
    per = np.diff(tzp)
    
    if np.sum(np.isfinite(per)) > 1:
        peravg = np.nanmean(per)
        perrms = np.nanstd(per)
    else:
        peravg = np.NaN
        perrms = np.NaN
    
    
    rotfavg = (1/peravg)
    rotfrms = rotfavg * perrms / peravg
    
    # make phase
    phase = np.NaN * np.ones(len(tss))
    for jper,_ in enumerate(per):
        j = np.where((tzp[jper]<tss) & (tss <= tzp[jper+1]))[0]
        phase[j] = 2*np.pi*(tss[j] - tzp[jper]) / per[jper]
    
    # sinusoidal fitting
    j=np.where(np.isfinite(phase))[0]
    nfit = len(j)
    
    #using matrix math to combine individual compass coil (direction) and EF (current speed) measurements for current
    #datapoint and calculate frequency amplitudes and phases for velocity estimates
    if nfit > len(phase)/2:
        aprxcc = np.stack([np.cos(phase[j]), np.sin(phase[j]), np.ones(len(j))])
        coefcc, _,_,_ = np.linalg.lstsq(aprxcc.T, fcss[j]) #coefcc = aprxcc \ fcss(j)
        rescc = fcss[j] - np.matmul(aprxcc.T, coefcc) # rescc = fcss[j] - aprxcc * coefcc
        fccr = np.nanstd(rescc)
        fcca = np.sqrt(coefcc[0]**2+coefcc[1]**2)
        fccp = np.arctan2(-coefcc[1],coefcc[0])
        ccbl = coefcc[2]
        
        aprxef = np.append(aprxcc, np.array([np.linspace(-1,1,len(j))]), axis=0) #aprxef = [aprxcc,  linspace(-1,1,length(j))']
        coefef, _,_,_ = np.linalg.lstsq(aprxef.T, fess[j]) #coefef = aprxef \ fess(j)'
        resef = fess[j] - np.matmul(aprxef.T, coefef) # resef = fess[j] - aprxef * coefef
        fefr = np.nanstd(resef)
        fefa = np.sqrt(coefef[0]**2+coefef[1]**2)
        fefp = np.arctan2(-coefef[1],coefef[0])
        efbl = coefef[2]
        
    else:
        ccbl = np.NaN
        efbl = np.NaN
        fccr = np.NaN
        fefr = np.NaN
        fcca = np.NaN
        fefa = np.NaN
        fccp = np.NaN
        fefp = np.NaN
        
    
    # probe gain and phase angles as function of rotation frequency
    gcca  = dataconvert(rotfavg,self.gcca_poly)
    gccp  = dataconvert(rotfavg,self.gccp_poly)
    gcora = dataconvert(rotfavg,self.gcora_poly)
    gcorp = dataconvert(rotfavg,self.gcorp_poly)
    gefa  = dataconvert(rotfavg,self.gefa_poly)
    gefp  = dataconvert(rotfavg,self.gefp_poly)
    
    # convert frequency amp and phase to velocity estimates
    vc0a = fcca / self.gcvfa / gcca * 1e6
    vc0p = fccp - self.gcvfp - gccp
    
    ve0a1 = fefa / self.gevfa / gefa * 1e6
    ve0p1 = fefp - self.gevfp - gefp
    
    ve0q1 = ve0a1 * np.cos(ve0p1)
    ve0i1 = ve0a1 * np.sin(ve0p1)
    
    ve0a2 = fcca / self.gcvfa / gcca * gcora / gefa * 1e6
    ve0p2 = fccp - self.gcvfp - gccp + gcorp - gefp
    
    ve0q2 = ve0a2 * np.cos(ve0p2)
    ve0i2 = ve0a2 * np.sin(ve0p2)
    
    ve0q = ve0q1 - ve0q2
    ve0i = ve0i1 - ve0i2
    
    ve0a = np.sqrt(ve0q**2+ve0i**2)
    ve0p = np.arctan2(ve0i,ve0q)
    
    if self.revcoil:
        ve0p += np.pi
        
    return rotfavg, rotfrms, efbl, ccbl, fefr, fccr, vc0a, vc0p, ve0a, ve0p, gcca, gefa
    
    
    
    
def calc_currents(self, rotfavg, fccr, fefr, vc0a, vc0p, ve0a, ve0p, gcca, gefa, nindep, w):

    #calculate current U/V components, velocity error, coil area/error
    area = vc0a / rotfavg * self.sfa
    aerr = np.abs(fccr/self.gcvfa/gcca*1e6/rotfavg*self.sfa)/np.sqrt(nindep)
    
    psi = -ve0p + vc0p + np.pi/2 + np.pi
    uw = w * (area/self.amean_rough) * self.sfw
    
    umag =  ve0a * np.cos(psi) * self.sfv
    vmag = -ve0a * np.sin(psi) * self.sfv + uw
    
    verr = np.abs(fefr * self.sfv / self.gevfa / gefa * 1e6) / np.sqrt(nindep)
    
    return area, aerr, umag, vmag, verr
    
    
    
    
    
    
    
    
    
    
    


#! /usr/bin/env/python3

# =============================================================================
#     Author: Casey R. Densmore, 12FEB2022
#
#    AXBPS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    AXBPS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with AXBPS.  If not, see <https://www.gnu.org/licenses/>.
# =============================================================================

import numpy as np
from scipy.io import wavfile #for wav file reading
from scipy.signal import tukey #taper generation

import wave #WAV file writing

import time as timemodule
import datetime as dt

from traceback import print_exc as trace_error

from shutil import copy as shcopy
from sys import getsizeof





# #initializes dict with default settings for AXCTD processor
# #any specified settings will be overwritten after this function is called in __init__
# def init_AXCP_settings(self, settings):

#     self.settings = {}
#     self.settings["refreshrate"] = 1 #size of raw audio chunks to process, in seconds
#     self.settings["revcoil"]     = False #coil on AXCP reversed- rotates currents by 180 degrees
#     self.settings["quality"]     = 1    #profile processing quality- 1=high/slow, 2=moderate speed/quality, 3=low/fast
#     self.settings["spindown_detect_rt"] = True #realtime detection of probe spindown to avoid processing unnecessary data
    
#     #1-use zero crossings and temperature baseline freq calculated with velocity
#     #2-use FFT (window size set by settings['tempfftwindow']) once per refresh
#     self.settings["temp_mode"] = 2 
#     self.settings["tempfftwindow"] = 1.0 #FFT window for temperature in seconds (used if temp_mode > 1) 
            
#     for csetting in settings:
#         self.settings[csetting] = settings[csetting] #overwrite defaults for user specified settings 
      
#     self.revcoil = bool(self.settings["revcoil"])
#     self.spindown_detect_rt = bool(self.settings["spindown_detect_rt"])
    
#     self.quality = self.settings["quality"]
#     if self.quality <= 0:
#         self.quality = 1
#     elif self.quality >= 4:
#         self.quality = 3
    
#     self.temp_mode = self.settings["temp_mode"]
#     if self.temp_mode < 1:
#         self.temp_mode = 1
#     elif self.temp_mode > 2:
#         self.temp_mode = 2    

    
#     self.refreshrate = self.settings["refreshrate"]
#     self.tempfftwindowsec = self.settings["tempfftwindow"]
    
#     if self.tempfftwindowsec > self.refreshrate: #temperature FFT window length must be less than refresh rate
#     self.tempfftwindowsec = self.refreshrate


    
# =============================================================================
#  Audio Processor class
# =============================================================================


class AXBT_Processor:

    #initializing current thread (saving variables, reading audio data or contacting/configuring receiver)
    def __init__(self, audiofile, audiochannel=-1, timerange=[0,-1], settings={}):
        
        self.audiochannel = audiochannel
        
        self.keepgoing = True  # signal connections
        self.waittoterminate = False #whether to pause on termination of run loop for kill process to complete
        
        self.timerange = timerange
        
        self.temperature = []
        self.depth = []
        self.time = []
        self.Rp = []
        self.Sp = []
        self.fp = []
        
        self.audiofile = audiofile
        
        self.init_settings(settings)
        
        
        
    def init_settings(self,settings):
        
        self.settings = {'fftwindow':0.3, 'minfftratio':0.5, 'minsiglev':65, 'triggerfftratio':0.88, 'triggersiglev':75, 'tcoeff':[-40.0,0.02778,0.0,0.0], 'zcoeff':[0.0,1.524,0.0,0.0], 'flims':[1300,2800]}
        
        #overwriting user-defined settings
        for csetting in settings:
            self.settings[csetting] = settings[csetting]
        
        #FFT thresholds
        self.fftwindow = self.settings['fftwindow']
        self.minfftratio = self.settings['minfftratio']
        self.minsiglev = self.settings['minsiglev']
        self.triggerfftratio = self.settings['triggerfftratio']
        self.triggersiglev = self.settings['triggersiglev']
        
        #conversion coefficients + parameters
        self.tcoeff = self.settings['tcoeff']
        self.zcoeff = self.settings['zcoeff']
        self.flims = self.settings['flims']
        
        
        
        
        
        
    def init_window(self,N):
        
        # apply taper- alpha=0.25
        self.taper = tukey(N, alpha=0.25)
        
        self.N = N
        
        T = N/self.f_s
        df = 1 / T
        self.f = np.array([df * n if n < N / 2 else df * (n - N) for n in range(N)])#constraining peak frequency options to frequencies in specified band
        self.good_f_ind = np.all((np.greater_equal(self.f, self.flims[0]), np.less_equal(self.f, self.flims[1])), axis=0)
        
        self.good_f = self.f[self.good_f_ind]
                
        
        
    
    #convert time to depth, freq to temp given coefficient lists
    def btconvert(self,input,coefficients):
        output = 0
        for (i,c) in enumerate(coefficients):
            output += c*input**i
        return output
        
        
        
    
    #function to run fft here
    def dofft(self,pcmdata):
        pcmdata = self.taper * pcmdata
    
        # conducting fft, converting to real space
        fftdata = np.abs(np.fft.fft(pcmdata))
        fftdata_inrange = fftdata[self.good_f_ind]
        
        maxind = np.argmax(fftdata_inrange)
        
        #frequency of max signal within band (AXBT-transmitted frequency)
        fp = self.good_f[maxind] 
        
        #maximum signal strength in band
        Sp = 10*np.log10(fftdata_inrange[maxind])
    
        #ratio of maximum signal in band to max signal total (SNR)
        Rp = fftdata_inrange[maxind]/np.max(fftdata) 
        
        # #Signal level normalized by power at quiet frequency
        # pdead = fftdata[self.dead_freq_ind]
        # Sp = np.log10(fftdata_inrange[maxind]/pdead)
            
        return fp, Sp, Rp
        
        
        
        
    def run(self):
        
        #checking file length- wont process files with more frames than max size
        try: #exception if unable to read audio file if it doesn't exist or isn't WAV formatted
            file_info = wave.open(self.audiofile)
        except:
            print("[!] Error- unable to open audio file! Terminating...")
            return
        
        self.f_s, snd = wavfile.read(self.audiofile) #reading file
        
        #if multiple channels, sum them together
        sndshape = np.shape(snd) #array size (tuple)
        ndims = len(sndshape) #number of dimensions
        if ndims == 1: #if one channel, use that
            self.audiostream = snd
            
        elif ndims == 2: #if multiple channels, pick selected channel, otherwise sum
            if self.audiochannel >= 1:
                if sndchape[1] > sndshape [0]:
                    self.audiostream = snd[self.audiochannel-1,:]
                else:
                    self.audiostream = snd[:,self.audiochannel-1]
                    
            else:
                self.audiostream = np.sum(snd,axis=1)
                print("[+] Multiple channels in file and channel not specified- using sum of all channels")
                
        else: #if more than 2D- not a valid file
            print("[!] Error- audio file has more than two dimensions! Terminating...")
            return
                
        #trimming audio as required
        if self.timerange[1] > 0: #trim end of file first to keep this simple
            e = int(round(self.f_s * self.timerange[1]))
            self.audiostream = self.audiostream[:e]
        if self.timerange[0] > 0:
            s = int(round(self.f_s * self.timerange[0]))
            self.audiostream = self.audiostream[s:]
            
        #configuring sample times for the audio file
        self.lensignal = len(self.audiostream)
        self.maxtime = self.lensignal/self.f_s
        self.sampletimes = np.arange(0.1,self.maxtime-0.1,0.1)
        self.ntimes = len(self.sampletimes)
        
        self.istriggered = False
        self.starttime = 0
            
        # setting up thread while loop- terminates when user clicks "STOP" or audio file finishes processing
        i = -1
            
        
        N_cur = -1
        
        #MAIN PROCESSOR LOOP
        while self.keepgoing:
            i += 1 

            #kill test/audio threads once time exceeds the max time of the audio file
            #NOTE: need to do this on the cycle before hitting the max time when processing from audio because
            #       the WAV file processes faster than the thread can kill itself
            if  i >= len(self.sampletimes)-1:
                self.keepgoing = False
                print(f"[+] Processing status: Completed!")
                return
                
            #getting current time to sample from audio file
            ctime = self.sampletimes[i]

            #getting current data to sample from audio file- using indices like this is much more efficient than calculating times and using logical arrays
            ctrind = int(np.round(ctime*self.f_s))
            pmind = int(np.min([np.round(self.f_s*self.fftwindow/2),ctrind,self.lensignal-ctrind-1])) #uses minimum value so no overflow
            currentdata = self.audiostream[ctrind-pmind:ctrind+pmind]
                
            #setting up frequency array as reqd
            N = len(currentdata)
            if N_cur != N:
                self.init_window(N)
                N_cur = N

            #conducting FFT or skipping, depending on signal strength
            fp,Sp,Rp = self.dofft(currentdata)        
    
            #rounding before comparisons happen
            ctime = np.round(ctime, 1)
            fp = np.round(fp, 2)
            Sp = np.round(Sp, 2)
            Rp = np.round(Rp, 3)        
            

                
            #logic to determine whether or not profile is triggered
            if not self.istriggered and Sp >= self.triggersiglev and Rp >= self.triggerfftratio:
                self.istriggered = True
                self.firstpointtime = ctime
                    
            #logic to determine whether or not point is valid
            if self.istriggered and Sp >= self.minsiglev and Rp >= self.minfftratio:
                cdepth = self.btconvert(ctime - self.firstpointtime, self.zcoeff)
                ctemp = self.btconvert(fp, self.tcoeff)
                
                #rounding
                ctemp = np.round(ctemp, 2)
                cdepth = np.round(cdepth, 1)
                
            else:
                ctemp = np.nan
                cdepth = np.nan
                
            #saving points to lists
            self.temperature.append(ctemp)
            self.depth.append(cdepth)
            self.time.append(ctime)
            self.Rp.append(Rp)
            self.Sp.append(Sp)
            self.fp.append(fp)
                
            
            print(f"[+] Processing status: {round(100*i/self.ntimes)}%         ", end='\r')
            timemodule.sleep(0.001) #slight pause to free some resources when processing from audio
                
            
        
        

        
        
        


# =============================================================================
#     Author: Casey R. Densmore, 12FEB2022
#
#    This file is part of the Airborne eXpendable Buoy Processing System (AXBPS)
#
#    AXBPS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    AXBPS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with AXBPS.  If not, see <https://www.gnu.org/licenses/>.
# =============================================================================

import numpy as np
from datetime import date, datetime, timedelta
import chardet

    
    
    
    
def writelogfile(logfile, dropdatetime, depth, tempC, frequency, time, probetype):
    
    header = f""" Probe Type = {probetype}     
       Date = {dropdatetime:%Y/%m/%d}
       Time = {dropdatetime:%H:%M:%S}
 
    Time     Depth    Frequency    (C)       (F) 
    """
    
    with open(logfile,'w') as f_out:
        f_out.write(header + "\n")
        
        for (ct,cf,cd,cT) in zip(time,frequency,depth,tempC):
            
            if np.isnan(cf*cT*cd) or cf <= 0: #invalid point
                cline = f'{ct:7.1f}{-10:10.1f}{0:11.1f}    ******    ******'
                
            else: #valid data point
                cTf = cT*9/5 + 32 #degF
                cline = f'{ct:7.1f}{cd:10.1f}{cf:11.1f}{cT:10.2f}{cTf:10.2f}'
                
            f_out.write(cline + "\n")
        
            



def writeedffile(edffile,dropdatetime,lat,lon,data,comments,QC=False):
    
    
    with open(edffile,'w') as f_out:
    
        #writing header, date and time, drop # (bad value)
        f_out.write("// This is an air-launched probe EXPORT DATA FILE  (EDF)\n")
        f_out.write("// File generated with the Airborne eXpendable Buoy Processing System (AXBPS)\n")
        f_out.write(f"Date of Launch:  {datetime.strftime(dropdatetime,'%m/%d/%y')}\n")
        f_out.write(f"Time of Launch:  {datetime.strftime(dropdatetime,'%H:%M:%S')}\n")
        
        #latitude and longitude in degrees + decimal minutes
        if lat >= 0:
            nsh = 'N'
        else:
            nsh = 'S'
        if lon >= 0:
            ewh = 'E'
        else:
            ewh = 'W'
        lon = abs(lon)
        lat = abs(lat)
        latdeg = int(np.floor(lat))
        londeg = int(np.floor(lon))
        latmin = (lat - latdeg)*60
        lonmin = (lon - londeg)*60
        f_out.write(f"Latitude      :  {latdeg:02d} {latmin:06.3f}{nsh}\n")
        f_out.write(f"Longitude     :  {londeg:03d} {lonmin:06.3f}{ewh}\n")        
        
        if QC:
            qcstr = " "
        else:
            qcstr = """
// This profile has not been quality-controlled.
//
"""
        
        #building drop metadata comments section
        drop_settings_info = f"""//
// Drop Settings Information:
""" + comments + """
""" + "\n".join([f"Field{i+1}  :  {ckey}" for i,ckey in enumerate(data.keys())]) + """
//""" + qcstr + """
""" + "\t".join([ckey for ckey in data.keys()]) + "\n"
        
        f_out.write(drop_settings_info)
        
        fields = list(data.keys())
        
        #determining what string format to use for each field
        field_formats = []
        for cfield in fields:
            cfieldlow = cfield.lower()
            if 'temperature' in cfieldlow or 'conductivity' in cfieldlow or 'salinity' in cfieldlow:
                field_formats.append('7.2f') #__XX.XX
            elif 'current' in cfieldlow:
                field_formats.append('6.2f') #__X.Xx
            elif 'depth' in cfieldlow or 'time' in cfieldlow or 'frequency' in cfieldlow:
                field_formats.append('9.2f') #__XXXX.XX
            else:
                field_formats.append('') #unspecified format
                
        npts = len(data[fields[0]])
        
        #writing data
        for i in range(npts):
            cline = "\t".join([f"{data[cfield][i]:{cformat}}" for cfield,cformat in zip(fields,field_formats)]) #tab-delimited
            f_out.write(cline + "\n")

    
            
            




#write data to FIN file
def writefinfile(finfile,cdtg,lat,lon,num,depth,temperature,salinity=None,U=None,V=None):
    
    hasSal = False
    hasCurrents = False
    if salinity is not None: 
        hasSal = True
    elif U is not None and V is not None:
        hasCurrent = True
    
        
    with open(finfile,'w') as f_out:
    
        dayofyear = date.toordinal(date(cdtg.year,cdtg.month,cdtg.day)) - date.toordinal(date(cdtg.year-1,12,31))

        #formatting latitude string
        if lat >= 0:
            latsign = ' '
        else:
            latsign = '-'
            lat = abs(lat)

        #formatting longitude string
        if lon >= 0:
            lonsign = ' '
        else:
            lonsign = '-'
            lon = abs(lon)

        #writing header data
        f_out.write(f"{cdtg.year}   {dayofyear:03d}   {datetime.strftime(cdtg,'%H%M')}   {latsign}{lat:06.3f}   {lonsign}{lon:07.3f}   {num:02d}   6   {len(depth)}   0   0   \n")
        
        #creating list of data points
        Npts = len(depth)
        outdata = []
        if hasSal:
            obsperline = 3
            nobtypes = 3
            for (t,d,s) in zip(temperature,depth,salinity):
                outdata.extend([f"{t: 8.3f}",f"{d: 8.1f}",f"{s: 8.3f}"])
        elif hasCurrents:
            obsperline = 3
            nobtypes = 4
            for (t,d,u,v) in zip(temperature,depth,U,V):
                outdata.extend([f"{t: 8.3f}",f"{d: 8.1f}",f"{u: 8.3f}",f"{v: 8.3f}"])
        else:
            nobtypes = 2
            obsperline = 5
            for (t,d) in zip(temperature,depth):
                outdata.extend([f"{t: 8.3f}",f"{d: 8.1f}"])
                
        pointsperline = nobtypes*obsperline

        #writing profile data
        i = 0
        while i < Npts:
            if i+pointsperline < Npts:
                pointstopull = pointsperline
            else:
                pointstopull = Npts - i
                
            line = "".join(outdata[i:i+pointstopull]) + "\n"
            f_out.write(line)
            i += pointstopull
            

            
            
            
            


            

        


# =============================================================================
#     Author: Casey R. Densmore, 12FEB2022
#
#    AXBPS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    AXBPS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with AXBPS.  If not, see <https://www.gnu.org/licenses/>.
# =============================================================================

# =============================================================================
#   This code is translated/updated from the MATLAB xcpdsp.m script written by 
#       John Dunlap, University of Washington APL, 4 November 2009
# =============================================================================

import os
import numpy as np
from scipy import signal
from ._AXCP_convert_fxns import calc_temp_from_freq, dataconvert


    


#initializes dict with default settings for AXCTD processor
#any specified settings will be overwritten after this function is called in __init__
def init_AXCP_settings(self, settings):

    self.settings = {}
    self.settings["refreshrate"] = 1 #size of raw audio chunks to process, in seconds
    self.settings["revcoil"]     = False #coil on AXCP reversed- rotates currents by 180 degrees
    self.settings["quality"]     = 1    #profile processing quality- 1=high/slow, 2=moderate speed/quality, 3=low/fast
    self.settings["spindown_detect_rt"] = True #realtime detection of probe spindown to avoid processing unnecessary data
    
    #1-use zero crossings and temperature baseline freq calculated with velocity
    #2-use FFT (window size set by settings['tempfftwindow']) once per refresh
    self.settings["temp_mode"] = 2 
    self.settings["tempfftwindow"] = 1.0 #FFT window for temperature in seconds (used if temp_mode > 1) 
            
    for csetting in settings:
        self.settings[csetting] = settings[csetting] #overwrite defaults for user specified settings 
      
    self.revcoil = bool(self.settings["revcoil"])
    self.spindown_detect_rt = bool(self.settings["spindown_detect_rt"])
    
    self.quality = self.settings["quality"]
    if self.quality <= 0:
        self.quality = 1
    elif self.quality >= 4:
        self.quality = 3
    
    self.temp_mode = self.settings["temp_mode"]
    if self.temp_mode < 1:
        self.temp_mode = 1
    elif self.temp_mode > 2:
        self.temp_mode = 2    

    
    self.refreshrate = self.settings["refreshrate"]
    self.tempfftwindowsec = self.settings["tempfftwindow"]
    
    if self.tempfftwindowsec > self.refreshrate: #temperature FFT window length must be less than refresh rate
        self.tempfftwindowsec = self.refreshrate



    
        
def initialize_AXCP_vars(self):
        
    self.tsamp = 1/self.f_s
    self.fnyq0 = self.f_s/2
    
    if self.quality == 1:
        self.nss1 =  1   # first subsampling
        self.nss2 = 200 # second subsampling
    elif self.quality == 2:
        self.nss1 =  3
        self.nss2 = 67
    elif self.quality == 3:
        self.nss1 =  5
        self.nss2 = 40
        
    #temperature FFT window calculation
    self.N_temp = int(np.round(self.f_s * self.tempfftwindowsec))
    self.init_fft_window(self.N_temp) #intialize window, frequencies, and taper 
                
    self.fnyq1 = self.fnyq0 / self.nss1
    self.fnyq2 = self.fnyq1 / self.nss2
    
    self.fzclp = 50
    
    # make input buffer size evenly divisible by all the subsampling
    # so filtering and subsampling is less complicated
    self.pointsperloop = np.round(self.f_s * self.refreshrate)
    self.pointsperloop = np.ceil(self.pointsperloop / self.nss1) * self.nss1
    self.pointsperloop = int(np.ceil(self.pointsperloop / self.nss1 / self.nss2) * self.nss1 * self.nss2)
    self.refreshrate = self.pointsperloop / self.f_s # actual seconds
    
    # first fit is from depth_beg to depth_beg + depth_chunk
    # next fit is depth_step deeper
    self.depth_beg = 7.5 
    self.depth_chunk = 5 
    self.depth_step  = 2
    
    #profile output arrays
    self.T       = np.array([])
    self.CCENV   = np.array([])
    self.PK      = np.array([])
    self.FCCDEV  = np.array([])
    self.FROTLP  = np.array([])
    self.FROTDEV = np.array([])
    self.TEMP_FFT = np.array([])
    self.DEPTH_FFT = np.array([])
    
    
    self.TIME = np.array([])
    self.DEPTH= np.array([])
    self.TEMP = np.array([])
    self.U_MAG    = np.array([])
    self.V_MAG    = np.array([])
    self.U_TRUE    = np.array([])
    self.V_TRUE    = np.array([])
    self.ROTF = np.array([])
    self.ROTFRMS = np.array([])
    self.AREA = np.array([])
    self.EFBL = np.array([])
    self.CCBL = np.array([])
    self.FEFR = np.array([])
    self.FCCR = np.array([])
    self.VERR = np.array([])
    self.AERR = np.array([])
    self.TERR = np.array([])
    self.W    = np.array([])
    self.ENVCC       = np.array([])
    self.ENVCCRMS    = np.array([])
    self.PEAK        = np.array([])
    self.FTBL = np.array([])
    self.VC0A = np.array([])
    self.VC0P = np.array([])
    self.VE0A = np.array([])
    self.VE0P = np.array([])
    self.GCCA = np.array([])
    self.GEFA = np.array([])
    self.NINDEP = np.array([])
    
    #change each iteration
    self.xccbpendkeep = 0
    self.xefbpendkeep = 0
    self.xtebpendkeep = 0
    self.fccbpendkeep = 0
    self.npp = 0  # input buffer no
    self.nff = 0  # least squares fit number
    
    self.tspinup = -1
    self.tspindown = -1
            
    #storing values of frequencies etc. for profile calculations
    self.pk = np.array([])
    self.envxcc = np.array([])
    self.fcc = np.array([])
    self.fef = np.array([])
    self.fte = np.array([])
    self.tim = np.array([])
    
    #initializing filter coefficients and indices
    self.init_filters()
    
    # initialize conversion polynomials
    self.init_constants()
            
    
    
def init_filters(self):
    
    
    self.sosxinlp = signal.butter(4,3000/self.fnyq0, btype='lowpass',output='sos')  # low pass before first subsampling
    
    #bandpasses for compass coil, EF (velocity), and temperature bands
    ccband = np.array([2000,2500])/self.fnyq1
    efband = np.array([1000,1500])/self.fnyq1
    tband = np.array([250,500])/self.fnyq1
    self.sosxccbp = signal.butter(4,ccband, btype='bandpass',output='sos') # (rotation)
    self.sosxefbp = signal.butter(4,efband, btype='bandpass',output='sos') # (current magnitude)
    self.sosxtebp = signal.butter(4,tband, btype='bandpass',output='sos') # (temperature)
    
    self.sosfzclp = signal.butter(4,self.fzclp/self.fnyq1, btype='lowpass',output='sos') # low pass zero crossing pulses
    self.sosenvxcclp = signal.butter(2,1/self.fnyq1, btype='lowpass',output='sos') #nyquist frequency low pass filter for subsampling first round
    
    cc2band = np.array([2,20])/self.fnyq2
    self.sosfccbp = signal.butter(2,cc2band, btype='bandpass',output='sos') #bandpass filter for fcc
    self.sosenvfcclp = signal.butter(2,1/self.fnyq2, btype='lowpass',output='sos') #lowpass pilter for fcc (fN)
    self.sosfrotlp = signal.butter(2,0.5/self.fnyq2, btype='lowpass',output='sos') #lowpass filter for rotation frequency
    
    frot2band = np.array([0.1,1.0])/self.fnyq2
    self.sosfrotbp = signal.butter(2,frot2band, btype='bandpass',output='sos') #bandpass filter for rotation frequency
    self.sosenvfrotlp = signal.butter(2,0.05/self.fnyq2, btype='lowpass',output='sos') 
    
    #z indices
    self.zxinlp = np.zeros((self.sosxinlp.shape[0], 2))
    
    self.zxccbp = np.zeros((self.sosxccbp.shape[0], 2))
    self.zxefbp = np.zeros((self.sosxefbp.shape[0], 2))
    self.zxtebp = np.zeros((self.sosxtebp.shape[0], 2))
    
    self.zfcclp = np.zeros((self.sosfzclp.shape[0], 2))
    self.zfeflp = np.zeros((self.sosfzclp.shape[0], 2))
    self.zftelp = np.zeros((self.sosfzclp.shape[0], 2))
    
    self.zenvxcclp = np.zeros((self.sosenvxcclp.shape[0], 2))
    
    self.zfccbp = np.zeros((self.sosfccbp.shape[0], 2))
    self.zenvfcclp = np.zeros((self.sosenvfcclp.shape[0], 2))
    self.zfrotlp = np.zeros((self.sosfrotlp.shape[0], 2))
    self.zfrotbp = np.zeros((self.sosfrotbp.shape[0], 2))
    self.zenvfrotlp = np.zeros((self.sosenvfrotlp.shape[0], 2))
    
    
    
def init_constants(self):
    # amplitude and phase angle polynomials
    self.gcca_poly  = [ 1809.877761,    -1.25653911,       0.18856556   ]
    self.gccp_poly  = [ 29.826971133,    10.265226263,    -0.176531452  ]
    self.gcora_poly = [ 898.89564739,    0.453188608,      0.0717425016 ]
    self.gcorp_poly = [ 56.552652832,    7.783581145,     -0.112472607  ]
    self.gefa_poly  = [ 23866.12044581,  107.99111272898, -0.905827321  ]
    self.gefp_poly  = [ -138.158938796,  9.759010754,     -0.180878978  ]
    
    #convert degrees to radians for phase gain coefficients
    self.gccp_poly = [np.pi/180*i for i in self.gccp_poly]
    self.gcorp_poly = [np.pi/180*i for i in self.gcorp_poly]
    self.gefp_poly = [np.pi/180*i for i in self.gefp_poly]
    
    # probe calibrations
    self.amean_rough = 580  # to match mendo.rt processing
    self.esep = 5.19
    self.c1 =   0.97
    self.c2 =  -0.02
    self.gevfa = 494.66 # to match mendo.rt processing
    self.gevfp = 0
    self.gcvfa = 500
    self.gcvfp = 0
    
    # temperature calibration
    self.tcalfreq  = [ 285.3,   449.1 ]  # cal frequency at 0 and 30 deg C points
    self.tcalres   = [ 16329,   4024  ]  # cal resistance at 0 and 30 deg C points
    self.tcal      = [ 1.73323e-3, 8.75509e-5, 1.64067e-5, -5.09882e-7 ]
    self.tcalrs = (self.tcalfreq[0]*self.tcalres[0]-self.tcalfreq[1]*self.tcalres[1]) / (self.tcalfreq[1]-self.tcalfreq[0])
    self.tcalcap = 1.0/(4.4*self.tcalfreq[0]*(self.tcalrs+self.tcalres[0]))
    self.tcor      = [-0.10, -7.5e-5]
    
    # velocity scaling
    self.sfv = 1.0/(self.fz*1e-5*self.esep*(1.0+self.c1))
    self.sfw = self.fh/self.fz*(1.0+self.c2)/(1.0+self.c1)
    self.sfa = 100.0/(2.0*np.pi*self.fh*1e-5)
    
    #depth scaling
    self.depth_poly =  [4.68, 4.377, -0.00044   ] # Prater, JTech, Dec 1991
    # self.inv_depth_poly = [3.11700848802192e-10,5.14227005928016e-06,0.228465938277901,-1.07390827937333]
    self.inv_depth_poly = [-1.07390827937333,0.228465938277901,5.14227005928016e-06, 3.11700848802192e-10]

    
    
    
    
    
    
    
def init_fft_window(self,N):
        
    # apply taper- alpha=0.25
    self.flims = [200,600]
    self.taper = signal.tukey(N, alpha=0.25)
    
    self.N_temp = N
    
    T = N/self.f_s
    df = 1 / T
    self.f = np.array([df * n if n < N / 2 else df * (n - N) for n in range(N)])#constraining peak frequency options to frequencies in specified band
    self.good_f_ind = np.all((np.greater_equal(self.f, self.flims[0]), np.less_equal(self.f, self.flims[1])), axis=0)
    
    self.good_f = self.f[self.good_f_ind]
    
    
    
#run fft to determine peak frequency in temperature band
def dofft(self,pcmdata):
    
    if self.N_temp != len(pcmdata): #correct window length and parameters if it's wrong for some reason
        self.N_temp = len(pcmdata)
        self.init_fft_window(self.N_temp)
    
    pcmdata = self.taper * pcmdata

    # conducting fft, converting to real space
    fftdata = np.abs(np.fft.fft(pcmdata))
    fftdata_inrange = fftdata[self.good_f_ind]
    
    maxind = np.argmax(fftdata_inrange)
    
    #frequency of max signal within band (AXBT-transmitted frequency)
    fp = self.good_f[maxind] 
    
    #maximum signal strength in band
    Sp = 10*np.log10(fftdata_inrange[maxind])

    #ratio of maximum signal in band to max signal total (SNR)
    Rp = fftdata_inrange[maxind]/np.max(fftdata) 
        
    return fp, Sp, Rp
    
    
    
    
    
    
            
def first_subsample(self, xinlp):
    
    # first subsampling
    xinlp = xinlp[self.nss1-1::self.nss1]
    
    # band pass filter to get just the three carrier frequencies
    xccbp,self.zxccbp = signal.sosfilt(self.sosxccbp,xinlp,zi=self.zxccbp)
    xefbp,self.zxefbp = signal.sosfilt(self.sosxefbp,xinlp,zi=self.zxefbp)
    xtebp,self.zxtebp = signal.sosfilt(self.sosxtebp,xinlp,zi=self.zxtebp)
    
    # envelope detect xccbp
    envxcclp,self.zenvxcclp = signal.sosfilt(self.sosenvxcclp,np.abs(xccbp),zi=self.zenvxcclp)
    
    
    ###  compass coil frequency
    xccbpendprev = self.xccbpendkeep # hard limit xccbp to get zero crossing pulses
    self.xccbpendkeep = xccbp[-1]
    x = np.append(xccbpendprev, xccbp)
    r = np.ones(len(x))
    r[x < 0] = -1
    xccbpzcp = np.abs(np.diff(r)) # zero crossing pulses
    
    # low pass zero crossing pulses to obtain fcc, freq of CC on wire
    fcclp,self.zfcclp = signal.sosfilt(self.sosfzclp,xccbpzcp,zi=self.zfcclp) 
    
    
    ###  EF (velocity) frequency 
    xefbpendprev = self.xefbpendkeep # hard limit xefbp to get zero crossing pulses
    self.xefbpendkeep = xefbp[-1]
    x = np.append(xefbpendprev, xefbp)
    r = np.ones(len(x))
    r[x < 0] = -1
    xefbpzcp = np.abs(np.diff(r))  # zero crossing pulses
    
    # low pass zero crossing pulses to obtain fef, freq of EF on wire
    feflp,self.zfeflp = signal.sosfilt(self.sosfzclp,xefbpzcp,zi=self.zfeflp) 
    
    
    #### temperature frequency
    xtebpendprev = self.xtebpendkeep # hard limit xtebp to get zero crossing pulses
    self.xtebpendkeep = xtebp[-1]
    x = np.append(xtebpendprev, xtebp)
    r = np.ones(len(x))
    r[x < 0] = -1
    xtebpzcp = np.abs(np.diff(r))  # zero crossing pulses
    
    # low pass zero crossing pulses to obtain fte, freq of Temp on wire
    ftelp,self.zftelp = signal.sosfilt(self.sosfzclp,xtebpzcp,zi=self.zftelp) 
    
    return envxcclp, fcclp, feflp, ftelp
    
    
    
    
    
    
def second_subsample(self, pklp, envxcclp, fcclp, feflp, ftelp):
    
    # second subsampling -- use to do LSQ fits, and use fcc2 for probe rotation detection
    pk_cur     = pklp[self.nss2-1::self.nss2]
    envxcc_cur = envxcclp[self.nss2-1::self.nss2]
    fcc_cur = fcclp[self.nss2-1::self.nss2] * 0.5 * self.fnyq1 #also correct frequency based on subsampling interval
    fef_cur = feflp[self.nss2-1::self.nss2] * 0.5 * self.fnyq1
    fte_cur = ftelp[self.nss2-1::self.nss2] * 0.5 * self.fnyq1
    dt_subsample = self.tsamp * self.nss1 * self.nss2
    n2 = len(fcc_cur) 
    tim_cur = np.linspace(self.T[-1]-(n2-1)*dt_subsample, self.T[-1], n2)
    
    # band pass fcc for two rotation detection methods below
    fccbp,self.zfccbp = signal.sosfilt(self.sosfccbp,fcc_cur,zi=self.zfccbp)
    
    # envelope detect fccbp
    envfcclp,self.zenvfcclp = signal.sosfilt(self.sosenvfcclp,np.abs(fccbp),zi=self.zenvfcclp)
    
    # hard limit fccbp in prep to get frotlp
    fccbpendprev = self.fccbpendkeep
    self.fccbpendkeep = fccbp[-1]
    x = np.append(fccbpendprev, fccbp)
    r = np.ones(len(x))
    r[x < 0] = -1
    fccbpzcp = np.abs(np.diff(r))
    
    # low pass signal.lfilter zero crossing pulses to estimate rotation frequency
    frotlp,self.zfrotlp = signal.sosfilt(self.sosfrotlp,fccbpzcp,zi=self.zfrotlp)
    frotlp = frotlp * 0.5 * self.fnyq2
    
    # get variability of frot bandpass then envelope detect
    frotbp,self.zfrotbp = signal.sosfilt(self.sosfrotbp,frotlp,zi=self.zfrotbp)
    
    # envelope detect frot, quiet frotdev means probe rotation rate is steady
    envfrotlp,self.zenvfrotlp = signal.sosfilt(self.sosenvfrotlp,np.abs(frotbp),zi=self.zenvfrotlp)
    
    return pk_cur, envxcc_cur, fcc_cur, fef_cur, fte_cur, tim_cur, envfcclp, frotlp, envfrotlp
    
    
    
    
    
    
    
def calc_current_datapoint(self, t1, t2):
    
    # select data for fitting
    current_inds = np.where((self.tim > t1) & (self.tim < t2))[0]
    tss  = self.tim[current_inds]
    ftss = self.fte[current_inds] #all temperature band peak frequencies in current range
    fess = self.fef[current_inds] #all EF (current speed) band peak freqs in current range
    fcss = self.fcc[current_inds] #all compass coil (direction/rotation) band peak freqs in current range
    envxccss = self.envxcc[current_inds]
    pkss     = self.pk[current_inds]
    
    tavg = np.nanmean(tss)
    # tbof = tss(end)
    
    tchunk = t2 - t1
    nindep = np.round(tchunk * self.fzclp)
    
    # depth & fall rate, w
    timd = tavg - self.tspinup
    depth =  dataconvert(timd,self.depth_poly)
    w     = - dataconvert(timd,[self.depth_poly[1],self.depth_poly[2]*2])
    
    # temperature frequency (mean)
    ftbl = np.nanmean(ftss)
    
    # frequency ftbl error for temperature
    x = np.arange(0,len(ftss)) #1:length(ftss)
    p = np.polyfit(x,ftss,deg=1) # linear fit
    r = ftss - np.polyval(p,x)
    ftbl_err = np.nanstd(r) / np.sqrt(nindep)
    
    
    #temperature calculation
    temp = self.calc_temp_from_freq(ftbl, depth)
    
    #temperature error calculation
    temp_off = self.calc_temp_from_freq(ftbl+ftbl_err, depth)
    terr = temp_off - temp
    
    
    #determining gain/phase info for current calculation
    rotfavg, rotfrms, efbl, ccbl, fefr, fccr, vc0a, vc0p, ve0a, ve0p, gcca, gefa = self.calc_vel_components(fcss, fess, tss)
        
    #calculate current U/V components, velocity error, coil area/error
    area, aerr, umag, vmag, verr = self.calc_currents(rotfavg, fccr, fefr, vc0a, vc0p, ve0a, ve0p, gcca, gefa, nindep, w)
    
    
    return tavg, depth, temp, umag, vmag, area, rotfavg, rotfrms, ftbl, efbl, ccbl, fefr, fccr, terr, verr, aerr, w, envxccss, pkss, vc0a, vc0p, ve0a, ve0p, gcca, gefa, nindep
        

                
                
#this function is called once per loop of the AXCP Processor and processes as much data as is available in the respective buffers 
def iterate_AXCP_process(self, e): 
    
    self.npp += 1 #iterate processor timestep datapoint counter
    
    self.T = np.append(self.T, e/self.f_s) #time at end of current PCM chunk
    self.PK = np.append(self.PK, np.max(np.abs(self.demod_buffer))) #peak audio value
    
    xinlp, self.zxinlp = signal.sosfilt(self.sosxinlp, self.demod_buffer, zi=self.zxinlp) #lowpass filter input buffer
    
    #running first subsample, applying filters, pulling big three frequency band zerocrossing points
    envxcclp, fcclp, feflp, ftelp = self.first_subsample(xinlp)
    
    pklp = self.PK[-1] * np.ones(len(envxcclp)) #peak value, length of first subsample
    self.CCENV = np.append(self.CCENV, envxcclp[-1] * np.sqrt(2)) #compass coil environment
    
    #running second subsample, pulling big three center frequencies for profile calculations
    pk_cur, envxcc_cur, fcc_cur, fef_cur, fte_cur, tim_cur, envfcclp, frotlp, envfrotlp = self.second_subsample(pklp, envxcclp, fcclp, feflp, ftelp)
    
    #saving rotation frequency info
    self.FCCDEV = np.append(self.FCCDEV, envfcclp[-1]*np.sqrt(2) )
    self.FROTLP = np.append(self.FROTLP, frotlp[-1] )
    self.FROTDEV = np.append(self.FROTDEV, envfrotlp[-1]*np.sqrt(2) )
    
    
    #only retain last 40 seconds from previous iterations, append on new values
    time_save = -40 #last 40 seconds retained
    retainind = np.where(self.tim > tim_cur[-1] - time_save)[0]
    if len(retainind) > 0:
        retainind = retainind[0]
    else:
        retainind = 0
    
    self.pk = np.append(self.pk[retainind:], pk_cur)
    self.envxcc = np.append(self.envxcc[retainind:], envxcc_cur)
    self.fcc = np.append(self.fcc[retainind:], fcc_cur)
    self.fef = np.append(self.fef[retainind:], fef_cur)
    self.fte = np.append(self.fte[retainind:], fte_cur)
    self.tim = np.append(self.tim[retainind:], tim_cur)
    
    #realtime spindown detection- avoid processing unnecessary data
    #finds last point where rotation frequency is 12-18 Hz and rotation RMS < 0.5j
    #checked before spinup check to avoid spinup/down detection on the same iteration for large chunk sizes
    if self.spindown_detect_rt and self.status and np.max(self.FROTLP[-10:]) >= 18 and np.min(self.FROTLP[-10:]) <= 12 and np.min(self.FROTDEV[-10:]) > 1:
        self.status = 0 #spun down
        self.keepgoing = False #stop processing new data, run spindown checks
        print(f"[+] Spindown (realtime) detected: {self.T[-1]:7.2f} seconds, cleaning up!")
    
    
    #checking to see if probe has spunup (update status to 1 if so)
    #requirements: time >= 5 seconds, rotation frequency deviation below 0.5 (steady state), rotation frequency between 10 and 20 Hz
    #precise spinup time is defined as achieving 8 Hz, half rotation rate of average 16 Hz spin
    #added self.tspinup requirement so it won't re spin up profiles that have been spun down by the realtime detector
    if not self.status and self.T[-1] >= 5 and self.FROTDEV[-1] <= 0.5 and 10 < self.FROTLP[-1] < 20 and self.tspinup < 0:
        
        self.status = 1 #noting profile has spun up,  finding precise spinup time
        
        #getting 0-centered, recent, good compass coil frequencies for 0-crossing analysis
        fcc0 = self.fcc - np.mean(np.array([cfcc for i,cfcc in enumerate(self.fcc) if self.tim[-1]-self.tim[i] <= 5 and not np.isnan(cfcc) and np.isfinite(cfcc)])) 
        r = [1 if cfcc0 >= 0 else -1 for cfcc0 in fcc0] #ID compass coil frequency zero crossings
        cross_points = np.where(np.diff(r) > 0)[0]
        
        #getting calculating frequency rate of change across zero crossing points to interpolate exact zero crossing times
        d_freq = fcc0[cross_points+1] - fcc0[cross_points]
        d_time = self.tim[cross_points+1] - self.tim[cross_points]
        tim_zc = self.tim[cross_points] - fcc0[cross_points] * d_time / d_freq #actual zerocross times
        rotf_zc = 1 / np.diff(tim_zc)
        
        #rotf_zc will start extremely high when there is no valid data, drop low for a second as the static compass coil frequency is detected, and then quickly increase above 8 Hz as the probe spins up. Spinup time is the first time after the point where the probe speed first exceeds 8 Hz
        spinup_ind = np.where(rotf_zc < 8)[0]
        if len(spinup_ind) > 0:
            try:
                self.tspinup = tim_zc[spinup_ind[-1] + 2]
            except IndexError:
                self.tspinup = tim_zc[-1]
        else:
            self.tspinup = tim_zc[0]
        
        print(f"[+] Spinup detected: {self.tspinup:7.2f} seconds")
        
    
    #handle updated temperature FFT calculation outside of self.status - grab most recent data    
    if self.temp_mode > 1:
        cpcm = self.demod_buffer[-self.N_temp:] #pull proper window length of most recent pcm data
        fp,_,_ = self.dofft(cpcm) #get peak frequency in temperature band via FFT
        
        if self.status:
            cdepth_fft = dataconvert(self.T[-1]-self.tspinup, self.depth_poly) #getting depth corresponding to most recent time
        else:
            cdepth_fft = -999 #leave depths as -999 until spinup time can be determined and used to process
            
        ctemp_fft = self.calc_temp_from_freq(fp,cdepth_fft) #convert peak frequency in temperature band to corresponding temperature
    else:
        ctemp_fft = np.NaN
        cdepth_fft = np.NaN
        
    self.TEMP_FFT = np.append(self.TEMP_FFT, ctemp_fft)
    self.DEPTH_FFT = np.append(self.DEPTH_FFT, cdepth_fft)
    
    
    #if spinup has been detected but depths haven't been filled in, do that
    if self.status and self.temp_mode > 1 and -999 in self.DEPTH_FFT:
        for i,d in enumerate(self.DEPTH_FFT):
            if d == -999:
                cz = dataconvert(self.T[i]-self.tspinup, self.depth_poly)
                self.DEPTH_FFT[i] = cz
    
            
    #if the profile is spun up- iterate through all times available to process profile datapoints
    if self.status: 
        
        process_new_depth = True
        while process_new_depth:
            
            d1 = self.depth_beg + self.nff*self.depth_step #starting depth of current segment
            t1 = dataconvert(d1, self.inv_depth_poly) + self.tspinup #starting time for current segment
            t2 = dataconvert(d1+self.depth_chunk, self.inv_depth_poly) + self.tspinup #ending time for current segment
                        
            if t2 > self.tim[-1]:
                process_new_depth = False #out of new data, wait to grab more AXCP data
            else:
                self.nff += 1 #iterate profile datapoint counter
                
                #calculate profile information for current datapoint
                tavg, depth, temp_zc, umag, vmag, area, rotfavg, rotfrms, ftbl, efbl, ccbl, fefr, fccr, terr, verr, aerr, w, envxccss, pkss, vc0a, vc0p, ve0a, ve0p, gcca, gefa, nindep = self.calc_current_datapoint(t1, t2)
                
                
                #determining which temperature to use: zero crossing or FFT (depends on self.temp_mode and how they match)
                if self.temp_mode == 1:
                    temp = temp_zc
                elif self.temp_mode == 2:
                    temp = np.interp(depth, self.DEPTH_FFT, self.TEMP_FFT)
                
                #saving current profile point
                self.PEAK = np.append(self.PEAK, np.max(pkss))
                
                self.TIME = np.append(self.TIME, tavg)
                self.DEPTH = np.append(self.DEPTH, depth)
                
                self.FTBL = np.append(self.FTBL, np.max(ftbl))
                self.TEMP = np.append(self.TEMP, temp)
                self.TERR = np.append(self.TERR, terr)
                
                self.U_MAG = np.append(self.U_MAG, umag)
                self.V_MAG = np.append(self.V_MAG, vmag)
                self.VERR = np.append(self.VERR, verr)
                
                self.ROTF = np.append(self.ROTF, rotfavg)
                self.ROTFRMS = np.append(self.ROTFRMS, rotfrms)
                self.AREA = np.append(self.AREA, area)
                self.AERR = np.append(self.AERR, aerr)
                
                self.EFBL = np.append(self.EFBL, efbl)
                self.CCBL = np.append(self.CCBL, ccbl)
                self.FEFR = np.append(self.FEFR, fefr)
                self.FCCR = np.append(self.FCCR, fccr)
                self.W = np.append(self.W, w)
                
                self.ENVCC = np.append(self.ENVCC, np.nanmean(envxccss))
                self.ENVCCRMS = np.append(self.ENVCCRMS, np.nanstd(envxccss))
                
                self.VC0A = np.append(self.VC0A, np.max(vc0a))
                self.VC0P = np.append(self.VC0P, np.max(vc0p))
                self.VE0A = np.append(self.VE0A, np.max(ve0a))
                self.VE0P = np.append(self.VE0P, np.max(ve0p))
                self.GCCA = np.append(self.GCCA, np.max(gcca))
                self.GEFA = np.append(self.GEFA, np.max(gefa))
                self.NINDEP = np.append(self.GEFA, np.max(nindep))
                
    
                

#refining the profile's spindown point/truncating, recalculating AMEAN, and updating velocity profile
#only call this function if the profile was spunup and valid data collected prior to spindown
def refine_spindown_prof(self):
    
    #finding updated spindown profile index nffspindown
    nffspindown = len(self.TIME) #default value wont truncate data
    goodpoints = np.where([1 if crotf > 12 and crotf < 18 and crotfrms < 0.5 else 0 for (crotf,crotfrms) in zip(self.ROTF, self.ROTFRMS)])[0] #good data- rotation rate 12-18 Hz with RMS error < 0.5
    nffspindown = goodpoints[-1] #new default value is last valid point meeting above criteria
    
    if len(goodpoints) > 0 and len(self.TIME) >= 20: #enough data to not cause buffer issues
        inds_refine = [i for i in range(goodpoints[-1]-10, goodpoints[-1]-1) if np.isfinite(self.ROTFRMS[i])]
        # inds_refine = inds_refine[np.where(np.isfinite(self.ROTFRMS[inds_refine]))[0]] #only indices with finite ROTFRMS values
        
        #need at least two values for this calculation to work
        #delete final point if ROTFRMS exceeds 2*RMS of previous few points
        if len(inds_refine) >= 2:
            rms = np.min([0.05, np.nanmean(self.ROTFRMS[inds_refine]) ]) #cap rms at 0.05
            avg = np.nanstd(self.ROTFRMS[inds_refine])
            
            if self.ROTFRMS[goodpoints[-1]] > avg + 2 * rms:
                nffspindown = goodpoints[-1] - 1
                
    self.tspindown = self.TIME[nffspindown]
                
    #truncating all profiles, converting to numpy arrays (already should be but just to be sure)
    self.TIME = np.asarray(self.TIME[:nffspindown])
    self.DEPTH = np.asarray(self.DEPTH[:nffspindown])
    self.TEMP = np.asarray(self.TEMP[:nffspindown])
    self.U_MAG = np.asarray(self.U_MAG[:nffspindown])
    self.V_MAG = np.asarray(self.V_MAG[:nffspindown])
    self.ROTF = np.asarray(self.ROTF[:nffspindown])
    self.ROTFRMS = np.asarray(self.ROTFRMS[:nffspindown])
    self.AREA = np.asarray(self.AREA[:nffspindown])
    self.EFBL = np.asarray(self.EFBL[:nffspindown])
    self.CCBL = np.asarray(self.CCBL[:nffspindown])
    self.FEFR = np.asarray(self.FEFR[:nffspindown])
    self.FCCR = np.asarray(self.FCCR[:nffspindown])
    self.VERR = np.asarray(self.VERR[:nffspindown])
    self.AERR = np.asarray(self.AERR[:nffspindown])
    self.TERR = np.asarray(self.TERR[:nffspindown])
    self.W = np.asarray(self.W[:nffspindown])
    self.ENVCC = np.asarray(self.ENVCC[:nffspindown])
    self.ENVCCRMS = np.asarray(self.ENVCCRMS[:nffspindown])
    self.PEAK = np.asarray(self.PEAK[:nffspindown])
    self.FTBL = np.asarray(self.FTBL[:nffspindown])
    self.VC0A = np.asarray(self.VC0A[:nffspindown])
    self.VC0P = np.asarray(self.VC0P[:nffspindown])
    self.VE0A = np.asarray(self.VE0A[:nffspindown])
    self.VE0P = np.asarray(self.VE0P[:nffspindown])
    self.GCCA = np.asarray(self.GCCA[:nffspindown])
    self.GEFA = np.asarray(self.GEFA[:nffspindown])
    self.NINDEP = np.asarray(self.NINDEP[:nffspindown])

    
    #updating AMEAN calculation
    good_areas = np.sort(self.AREA[np.where((np.isfinite(self.AREA)) & (self.AERR < 5))[0]])
    if len(good_areas) >= 10:
        inset = int(np.ceil(0.1 * len(good_areas))) #ignore outer +/- 10% of areas in distribution
        self.amean_calc = np.nanmean(good_areas[inset:-inset])
        
        #updating profile meridional current velocities
        self.V_MAG = self.V_MAG - self.sfw * self.W * self.AREA * (1/self.amean_rough - 1/self.amean_calc)
        
    
    
        
        
def calculate_true_velocities(self):
    
    #convert current profile from degrees magnetic to true
    vel = np.sqrt(self.U_MAG**2 + self.V_MAG**2)
    curdir_true = (np.pi/180) * (90 - (180/np.pi) * np.arctan2(self.V_MAG, self.U_MAG) + self.dec) #MEAT: magnetic -> east add -> true
    self.U_TRUE = vel * np.sin(curdir_true)
    self.V_TRUE = vel * np.cos(curdir_true)
    
    
    
    
    
        
    


# =============================================================================
#     Author: Casey R. Densmore, 12FEB2022
#
#    This file is part of the Airborne eXpendable Buoy Processing System (AXBPS)
#
#    AXBPS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    AXBPS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with AXBPS.  If not, see <https://www.gnu.org/licenses/>.
# =============================================================================


###################################################################################
#                                    IMPORTS                                      #
###################################################################################

import os
import sys
from collections import namedtuple


import numpy as np

import gsw
    


###################################################################################
#                               BITSTREAM PARSING                                 #
###################################################################################

# hexframes,times,depths,temps,conds,psals,next_buffer_ind = parse.parse_bitstream_to_profile(self.binary_buffer, binbufftimes, self.r7500_buffer, self.masks)
def parse_bitstream_to_profile(bitstream, times, r400_in, r7500_in, tempLUT, tcoeff, ccoeff, zcoeff):
    
    hexframes = [] # hexadecimal representation of frame
    proftime = [] #time (post-profile start) corresponding to each ob
    z = [] #depth
    T = [] #temperature
    C = [] #conductivity
    S = [] #salinity
    r400 = []
    r7500 = []
    
    #initializing fields for loop
    s = 0 #starting bit
    numbits = len(bitstream)
    
    #looping through bits until finished parsing
    while s < numbits - 32:
        
        foundMatch = False
        
        if s >= numbits - 32: #won't overrun bitstream
            break
            
        #pulling current segment
        frame = bitstream[s:s+32]
        
        #verifying that frame meets requirements, otherwise increasing start bit and rechecking
        if frame[0:2] != [1, 0] or not check_crc(frame) or not r7500_in[s] > 0:
            s += 1
            
        else: #good profile point
            ctime = times[s] #current profile time
            
            #converting frame to T/C/S/z
            Tint, Cint = convertFrameToInt(frame)
            cT, cC, cS, cz = convertIntsToFloats(Tint, Cint, ctime, tempLUT, tcoeff, ccoeff, zcoeff)
            
            #storing frame/time/profile data
            hexframes.append(binListToHex(frame))
            proftime.append(ctime)
            z.append(cz)
            T.append(cT)
            C.append(cC)
            S.append(cS)
            
            r400.append(r400_in[s])
            r7500.append(r7500_in[s])
            
            s += 32 #increase start bit by 32 to search for next frame

    # End parse bitstream
    return hexframes, proftime, z, T, C, S, r400, r7500, s
    

    
    
    

###################################################################################
#          FRAME CONVERSION TO TEMPERATURE/CONDUCTIVITY/SALINITY/DEPTH            #
###################################################################################

def convertFrameToInt(frame):
    """ Convert a frame to integer fields
    frame is a list of bits """
    Tint = binListToInt(frame[14:26])
    Cint = binListToInt(frame[2:14])
    
    return Tint, Cint
    

    
def convertIntsToFloats(Tint, Cint, time, tempLUT, tcoeff, ccoeff, zcoeff):
    """ Convert a list of integer data fields to observations (floats) """
    
    #depth from time
    z = dataconvert(time, zcoeff)
    
    #uncalibrated temperature and conductivity from integers
    if Tint >= 0 and Tint <= len(tempLUT)-1:
        Tuncal = tempLUT[Tint]
    else:
        Tuncal = np.NaN
        
    Cuncal = Cint * 60 / 4096
    
    #calibrated temperature and conductivity from uncalibrated
    T = dataconvert(Tuncal, tcoeff)
    C = dataconvert(Cuncal, ccoeff)
    
    #salinity from temperature/conductivity/depth
    S = gsw.SP_from_C(C,T,z) #assumes pressure (mbar) approx. equals depth (m)
        
    return T, C, S, z
        
        
        
        
def read_temp_LUT(filename):
    tempLUT = []
    with open(filename) as f:
        for line in f.readlines():
            cline = line.strip().split(',')
            if len(cline) >= 2:
                tempLUT.append(float(cline[1]))
                
    return tempLUT
    
    
    
###################################################################################
#                                 HEADER PARSING                                  #
###################################################################################   
    
    
    
def trim_header(bits_in):
    
    bits = bits_in.copy()
    bits[:25] = [True for _ in range(25)]
    
    last_index_pulse = 0 #index of last 1 bit preceded by 7 other 1 bits
    n_ones_in_last_25 = 0 #number of bit 1 in the last 25 bits
    
    for i,b in enumerate(bits):
        
        if b: #counting number of 1 bits in last 25 bits
            n_ones_in_last_25 += 1
            if i > 10:
                if np.sum(bits[i-7:i+1]) == 8: #if last 8 bits are all 1
                    last_index_pulse = i #save index
                    
        if i > 24: #first 25 bits (indices 0 to 24) are all ones
            if bits[i-25]: #subtract a bit from the total if the bit on the backside of the sliding window is a 1
                n_ones_in_last_25 -= 1
                
            if i >= 400 and n_ones_in_last_25 <= 20: # >0.5 sec into header and >1/5 of last 25 bits =0, PULSE END
                break
    
                
    header_bits = bits[last_index_pulse : last_index_pulse + 32*75] #pad with an extra three frames worth of bits to parse
    
    return header_bits
    
    
    
def initialize_axctd_metadata():
    
    #dict containing metadata (conversions, other header info) for the AXCTD
    axctd_metadata = {'tcoeff': [0,1,0,0], 'ccoeff': [0,1,0,0], 'zcoeff': [1,1,1,1], 'serial_no':None, 'probe_code':None, 'max_depth':None, 'misc':None, 'tcoeff_hex':['','','',''], 'ccoeff_hex':['','','',''], 'zcoeff_hex':['','','',''], 'tcoeff_valid':[False] * 4, 'ccoeff_valid':[False] * 4, 'zcoeff_valid':[False] * 4}
    
    return axctd_metadata
    
    
    
    
def parse_header(bits):
    
    #data format:
    #bits 0-1 = '10'
    #bits 2-9 = counter (increments 0-63 then '11111' followed by counter 0-7)
    #bits 10-25 = data 
    #bits 26-31 = CRC
    
    #logical list for whether each frame was detected
    counter_found = [False] * 72
    
    #whether or not conversion coefficients were successfully pulled for T/C/z
    tcoeff_valid = [False] * 4
    ccoeff_valid = [False] * 4
    zcoeff_valid = [False] * 4
    
    #dict containing metadata (conversions, other header info) for the AXCTD
    axctd_metadata = initialize_axctd_metadata()

    #parsing header into (up to) 72 frames
    frame_data = [None] * 72
    lastframe = -1
    Nbits = len(bits)
    s = 0
    
    past_seven = False
    while lastframe < 71 and s < Nbits-32:
        if bits[s:s+2] != [1, 0] or not check_crc(bits[s:s+32]):
            s += 1
            
        else: #frame
        
            #get counter, acknowledge it's found
            counter_bits = bits[s+2:s+10]
            if counter_bits[:5] == [1,1,1,1,1]:
                curcounter = binListToInt(counter_bits[5:]) + 64 #last 8 frames have counter 0-7 preceded by 11111
            else:
                curcounter = binListToInt(counter_bits) #first 64 frames just have counter
                
            if curcounter <= 71:
                counter_found[curcounter] = True
                lastframe = curcounter
                
                #convert frame data to hexidecimal format
                frame_data[curcounter] = binListToHex(bits[s+10:s+26])
            
            s += 32
    
            
    
    #pulling frame data into metadata dict
    if sum(counter_found[4:6]) == 2:
        axctd_metadata['serial_no'] = frame_data[4] + frame_data[5]
    
    if counter_found[6]: #frame 6: probe max depth
        axctd_metadata['max_depth'] = frame_data[6]
        
    if counter_found[7]: #frame 7: probe type (AXCTD should be A000)
        axctd_metadata['probe_code'] = frame_data[7]
    
    #frames 12-23: depth coefficients D3,D2,D1,D0 in 3-frame sequences
    for i,cf in enumerate(range(21,11,-3)): 
        if sum(counter_found[cf:cf+3]) == 3:
            axctd_metadata['zcoeff_hex'][i] = ''.join(frame_data[cf:cf+3])
    
    #frames 24-35: temp coefficients D3,D2,D1,D0 in 3-frame sequences
    for i,cf in enumerate(range(33,23,-3)): 
        if sum(counter_found[cf:cf+3]) == 3:
            axctd_metadata['tcoeff_hex'][i] = ''.join(frame_data[cf:cf+3])
            
    #frames 36-47: cond. coefficients D3,D2,D1,D0 in 3-frame sequences
    for i,cf in enumerate(range(45,35,-3)): 
        if sum(counter_found[cf:cf+3]) == 3:
            axctd_metadata['ccoeff_hex'][i] = ''.join(frame_data[cf:cf+3])
    
    #calculating conversion coefficients
    coeff_types = ['t','c','z']
    for coeff in coeff_types:
        for i in range(4):
            if axctd_metadata[coeff+'coeff_hex'][i] != '':
                chex = axctd_metadata[coeff+'coeff_hex'][i].upper().replace('B','+').replace('D','-')
                axctd_metadata[coeff+'coeff'][i] = int(chex[:9])/1E7 * 10**int(chex[9:])
                axctd_metadata[coeff+'coeff_valid'][i] = True
                
    #saving frame data to metadata
    axctd_metadata['frame_data'] = frame_data
    axctd_metadata['counter_found'] = counter_found
    
    return axctd_metadata
    
    
    
    


###################################################################################
#                   GENERAL LIST FORMAT DATA CONVERSION                           #
###################################################################################
    
#conversion: coefficients=C,  D_out = C[0] + C[1]*D_in + C[2]*D_in^2 + C[3]*D_in^3 + ...
def dataconvert(data_in,coefficients):
    output = 0
    for (i,c) in enumerate(coefficients):
        output += c*data_in**i
    return output
    
    


###################################################################################
#                   CYCLIC REDUNCANCY CHECK FOR FRAME                             #
###################################################################################

def check_crc(bits):
    
    divisor = [1,1,0,0,1,0,1]
    result = bits.copy()
    
    for k in range(26):
        if result[k]:
            for i in range(7):
                result[i+k] = result[i+k] != divisor[i]
            
    isGood = not sum(result) #returns 1 if sum is 0, otherwise 0
    
    return isGood

    

###################################################################################
#                         BINARY LIST / INTEGER CONVERSION                       #
###################################################################################


def binListToInt(binary):
    """ Convert a list of bits into a binary number """
    x = 0
    mask = 0x1 << (len(binary) - 1)
    for b in binary:
        if b:
            x |= mask
        mask >>= 1

    return x


def intToBinList(cInt, masklen):
    """ Convert a number into a list of bits with length
    at least masklen  """
    x = cInt
    i = 0
    bin_list = [0] * masklen
    while x:
        try:
            bin_list[i] = x & 1
        except IndexError: # if there are more bits than masklen
            bin_list.append(x & 1)
        x >>= 1
        i += 1

    bin_list.reverse()
    return bin_list

    
    
    
def binListToHex(binary):
    prof_starts = [i for i in range(0,len(binary),4)]
    prof_ends = [i for i in range(4,len(binary)+1,4)]
    
    letters = 'abcdef'
    
    hex_str = ''
    for s,e in zip(prof_starts, prof_ends):
        cbinnum = binListToInt(binary[s:e])
        if cbinnum <= 9: #convert decimal integer to hex digit
            cstr = str(cbinnum)
        else:
            cstr = letters[cbinnum-10]
            
        hex_str += cstr #append current hex digit to hex string
        
    return hex_str
    
    
    
            
